import os
import json
import mock

import pytest
from chalice.config import Config
from chalice import package
from chalice.constants import LAMBDA_TRUST_POLICY
from chalice.deploy.appgraph import ApplicationGraphBuilder, DependencyBuilder
from chalice.awsclient import TypedAWSClient
from chalice.deploy.deployer import BuildStage
from chalice.deploy import models
from chalice.deploy.swagger import SwaggerGenerator
from chalice.package import PackageOptions
from chalice.utils import OSUtils


@pytest.fixture
def mock_swagger_generator():
    return mock.Mock(spec=SwaggerGenerator)


def test_can_create_app_packager():
    config = Config()
    options = PackageOptions(mock.Mock(spec=TypedAWSClient))
    packager = package.create_app_packager(config, options)
    assert isinstance(packager, package.AppPackager)


def test_can_create_terraform_app_packager():
    config = Config()
    options = PackageOptions(mock.Mock(spec=TypedAWSClient))
    packager = package.create_app_packager(config, options,  'terraform')
    assert isinstance(packager, package.AppPackager)


def test_template_post_processor_moves_files_once():
    mock_osutils = mock.Mock(spec=OSUtils)
    p = package.SAMCodeLocationPostProcessor(mock_osutils)
    template = {
        'Resources': {
            'foo': {
                'Type': 'AWS::Serverless::Function',
                'Properties': {
                    'CodeUri': 'old-dir.zip',
                }
            },
            'bar': {
                'Type': 'AWS::Serverless::Function',
                'Properties': {
                    'CodeUri': 'old-dir.zip',
                }
            },
        }
    }
    p.process(template, config=None,
              outdir='outdir', chalice_stage_name='dev')
    mock_osutils.copy.assert_called_with(
        'old-dir.zip', os.path.join('outdir', 'deployment.zip'))
    assert mock_osutils.copy.call_count == 1
    assert template['Resources']['foo']['Properties']['CodeUri'] == (
        './deployment.zip'
    )
    assert template['Resources']['bar']['Properties']['CodeUri'] == (
        './deployment.zip'
    )


def test_terraform_post_processor_moves_files_once():
    mock_osutils = mock.Mock(spec=OSUtils)
    p = package.TerraformCodeLocationPostProcessor(mock_osutils)
    template = {
        'resource': {
            'aws_lambda_function': {
                'foo': {'filename': 'old-dir.zip'},
                'bar': {'filename': 'old-dir.zip'},
            }
        }
    }

    p.process(template, config=None,
              outdir='outdir', chalice_stage_name='dev')
    mock_osutils.copy.assert_called_with(
        'old-dir.zip', os.path.join('outdir', 'deployment.zip'))
    assert mock_osutils.copy.call_count == 1
    assert template['resource']['aws_lambda_function'][
        'foo']['filename'] == ('${path.module}/deployment.zip')
    assert template['resource']['aws_lambda_function'][
        'bar']['filename'] == ('${path.module}/deployment.zip')


def test_template_generator_default():
    tgen = package.TemplateGenerator(Config(),
                                     PackageOptions(
                                         mock.Mock(spec=TypedAWSClient)
                                     ))

    with pytest.raises(package.UnsupportedFeatureError):
        tgen.dispatch(models.Model(), {})


class TestTemplateMergePostProcessor(object):
    def _test_can_call_merge(self, file_template, template_name):
        mock_osutils = mock.Mock(spec=OSUtils)
        mock_osutils.get_file_contents.return_value = json.dumps(file_template)
        mock_merger = mock.Mock(spec=package.TemplateMerger)
        mock_merger.merge.return_value = {}
        p = package.TemplateMergePostProcessor(
            mock_osutils, mock_merger, package.JSONTemplateSerializer(),
            merge_template=template_name)
        template = {
            'Resources': {
                'foo': {
                    'Type': 'AWS::Serverless::Function',
                    'Properties': {
                        'CodeUri': 'old-dir.zip',
                    }
                },
                'bar': {
                    'Type': 'AWS::Serverless::Function',
                    'Properties': {
                        'CodeUri': 'old-dir.zip',
                    }
                },
            }
        }

        config = mock.MagicMock(spec=Config)

        p.process(
            template, config=config, outdir='outdir', chalice_stage_name='dev')

        assert mock_osutils.file_exists.call_count == 1
        assert mock_osutils.get_file_contents.call_count == 1
        mock_merger.merge.assert_called_once_with(file_template, template)

    def test_can_call_merge(self):
        file_template = {
            "Resources": {
                "foo": {
                    "Properties": {
                        "Environment": {
                            "Variables": {"Name": "Foo"}
                        }
                    }
                }
            }
        }

        self._test_can_call_merge(file_template, 'extras.json')

    def test_can_call_merge_with_yaml(self):
        file_template = '''
            Resources:
              foo:
                Properties:
                  Environment:
                    Variables:
                      Name: Foo
        '''

        self._test_can_call_merge(file_template, 'extras.yaml')

    def test_raise_on_bad_json(self):
        mock_osutils = mock.Mock(spec=OSUtils)
        mock_osutils.get_file_contents.return_value = (
            '{'
            '  "Resources": {'
            '    "foo": {'
            '      "Properties": {'
            '        "Environment": {'
            '          "Variables": {"Name": "Foo"}'
            ''
        )
        mock_merger = mock.Mock(spec=package.TemplateMerger)
        p = package.TemplateMergePostProcessor(
            mock_osutils, mock_merger, package.JSONTemplateSerializer(),
            merge_template='extras.json')
        template = {}

        config = mock.MagicMock(spec=Config)
        with pytest.raises(RuntimeError) as e:
            p.process(
                template,
                config=config,
                outdir='outdir',
                chalice_stage_name='dev',
            )
        assert str(e.value).startswith('Expected')
        assert 'to be valid JSON template' in str(e.value)
        assert mock_merger.merge.call_count == 0

    def test_raise_on_bad_yaml(self):
        mock_osutils = mock.Mock(spec=OSUtils)
        mock_osutils.get_file_contents.return_value = (
            '---'
            'Resources:'
            '    foo:'
            '      Properties:'
            '        Environment:'
            '          - 123'
            ''
        )
        mock_merger = mock.Mock(spec=package.TemplateMerger)
        p = package.TemplateMergePostProcessor(
            mock_osutils, mock_merger, package.YAMLTemplateSerializer(),
            merge_template='extras.yaml')
        template = {}

        config = mock.MagicMock(spec=Config)
        with pytest.raises(RuntimeError) as e:
            p.process(
                template,
                config=config,
                outdir='outdir',
                chalice_stage_name='dev',
            )
        assert str(e.value).startswith('Expected')
        assert 'to be valid YAML template' in str(e.value)
        assert mock_merger.merge.call_count == 0

    def test_raise_if_file_does_not_exist(self):
        mock_osutils = mock.Mock(spec=OSUtils)
        mock_osutils.file_exists.return_value = False
        mock_merger = mock.Mock(spec=package.TemplateMerger)
        p = package.TemplateMergePostProcessor(
            mock_osutils, mock_merger, package.JSONTemplateSerializer(),
            merge_template='extras.json')
        template = {}

        config = mock.MagicMock(spec=Config)
        with pytest.raises(RuntimeError) as e:
            p.process(
                template,
                config=config,
                outdir='outdir',
                chalice_stage_name='dev',
            )
        assert str(e.value).startswith('Cannot find template file:')
        assert mock_merger.merge.call_count == 0


class TestCompositePostProcessor(object):
    def test_can_call_no_processors(self):
        processor = package.CompositePostProcessor([])
        template = {}
        config = mock.MagicMock(spec=Config)
        processor.process(template, config, 'out', 'dev')

        assert template == {}

    def test_does_call_processors_once(self):
        mock_processor_a = mock.Mock(spec=package.TemplatePostProcessor)
        mock_processor_b = mock.Mock(spec=package.TemplatePostProcessor)
        processor = package.CompositePostProcessor(
            [mock_processor_a, mock_processor_b])
        template = {}
        config = mock.MagicMock(spec=Config)
        processor.process(template, config, 'out', 'dev')

        mock_processor_a.process.assert_called_once_with(
            template, config, 'out', 'dev')
        mock_processor_b.process.assert_called_once_with(
            template, config, 'out', 'dev')


class TemplateTestBase(object):

    template_gen_factory = None

    def setup_method(self, stubbed_session):
        self.resource_builder = package.ResourceBuilder(
            application_builder=ApplicationGraphBuilder(),
            deps_builder=DependencyBuilder(),
            build_stage=mock.Mock(spec=BuildStage)
        )
        client = TypedAWSClient(None)
        m_client = mock.Mock(wraps=client, spec=TypedAWSClient)
        type(m_client).region_name = mock.PropertyMock(
            return_value='us-west-2')
        self.pkg_options = PackageOptions(m_client)
        self.template_gen = self.template_gen_factory(
            Config(), self.pkg_options)

    def generate_template(self, config, chalice_stage_name='dev',
                          options=None):
        resources = self.resource_builder.construct_resources(
            config, chalice_stage_name)
        if options is None:
            options = self.pkg_options
        return self.template_gen_factory(config, options).generate(resources)

    def lambda_function(self):
        return models.LambdaFunction(
            resource_name='foo',
            function_name='app-dev-foo',
            environment_variables={},
            runtime='python27',
            handler='app.app',
            tags={'foo': 'bar'},
            timeout=120,
            xray=None,
            memory_size=128,
            deployment_package=models.DeploymentPackage(filename='foo.zip'),
            role=models.PreCreatedIAMRole(role_arn='role:arn'),
            security_group_ids=[],
            subnet_ids=[],
            layers=[],
            reserved_concurrency=None,
        )

    def managed_layer(self):
        return models.LambdaLayer(
            resource_name='layer',
            layer_name='bar',
            runtime='python2.7',
            deployment_package=models.DeploymentPackage(filename='layer.zip')
        )


class TestPackageOptions(object):

    def test_service_principal(self):
        awsclient = mock.Mock(spec=TypedAWSClient)
        awsclient.region_name = 'us-east-1'
        awsclient.endpoint_dns_suffix.return_value = 'amazonaws.com'
        awsclient.service_principal.return_value = 'lambda.amazonaws.com'
        options = package.PackageOptions(awsclient)
        principal = options.service_principal('lambda')
        assert principal == 'lambda.amazonaws.com'

        awsclient.endpoint_dns_suffix.assert_called_once_with('lambda',
                                                              'us-east-1')
        awsclient.service_principal.assert_called_once_with('lambda',
                                                            'us-east-1',
                                                            'amazonaws.com')


class TestTerraformTemplate(TemplateTestBase):

    template_gen_factory = package.TerraformGenerator

    EmptyPolicy = {
        'Version': '2012-10-18',
        'Statement': {
            'Sid': '',
            'Effect': 'Allow',
            'Action': 'lambda:*'
        }
    }

    def generate_template(self, config, chalice_stage_name='dev',
                          options=None):
        resources = self.resource_builder.construct_resources(
            config, chalice_stage_name)

        # Patch up resources that have mocks (due to build stage)
        # that we need to serialize to json.
        for r in resources:
            # For terraform rest api construction, we need a swagger
            # doc on the api resource as we'll be serializing it to
            # json.
            if isinstance(r, models.RestAPI):
                r.swagger_doc = {
                    'info': {'title': 'some-app'},
                    'x-amazon-apigateway-binary-media-types': []
                }
            if (isinstance(r, models.RestAPI) and
                    config.api_gateway_endpoint_type == 'PRIVATE'):
                r.swagger_doc['x-amazon-apigateway-policy'] = (
                    r.policy.document)

            # Same for iam policies on roles
            elif isinstance(r, models.FileBasedIAMPolicy):
                r.document = self.EmptyPolicy

        if options is None:
            options = self.pkg_options
        return self.template_gen_factory(config, options).generate(resources)

    def get_function(self, template):
        functions = list(template['resource'][
            'aws_lambda_function'].values())
        assert len(functions) == 1
        return functions[0]

    def test_supports_precreated_role(self):
        builder = DependencyBuilder()
        resources = builder.build_dependencies(
            models.Application(
                stage='dev',
                resources=[self.lambda_function()],
            )
        )
        template = self.template_gen.generate(resources)
        assert template['resource'][
            'aws_lambda_function']['foo']['role'] == 'role:arn'

    def test_adds_env_vars_when_provided(self, sample_app):
        function = self.lambda_function()
        function.environment_variables = {'foo': 'bar'}
        template = self.template_gen.generate([function])
        tf_resource = self.get_function(template)
        assert tf_resource['environment'] == {
            'variables': {
                'foo': 'bar'
            }
        }

    def test_adds_vpc_config_when_provided(self):
        function = self.lambda_function()
        function.security_group_ids = ['sg1', 'sg2']
        function.subnet_ids = ['sn1', 'sn2']
        template = self.template_gen.generate([function])
        tf_resource = self.get_function(template)
        assert tf_resource['vpc_config'] == {
            'subnet_ids': ['sn1', 'sn2'],
            'security_group_ids': ['sg1', 'sg2']}

    def test_adds_layers_when_provided(self):
        function = self.lambda_function()
        function.layers = layers = ['arn://layer1', 'arn://layer2']
        template = self.template_gen.generate([function])
        tf_resource = self.get_function(template)
        assert tf_resource['layers'] == layers

    def test_adds_managed_layer_when_provided(self):
        function = self.lambda_function()
        function.layers = ['arn://layer1', 'arn://layer2']
        function.managed_layer = self.managed_layer()
        template = self.template_gen.generate(
            [function.managed_layer, function])
        tf_resource = self.get_function(template)
        assert tf_resource['layers'] == [
            '${aws_lambda_layer_version.layer.arn}',
            'arn://layer1',
            'arn://layer2',
        ]
        assert template['resource']['aws_lambda_layer_version']['layer'] == {
            'layer_name': 'bar',
            'compatible_runtimes': ['python2.7'],
            'filename': 'layer.zip',
        }

    def test_adds_reserved_concurrency_when_provided(self, sample_app):
        function = self.lambda_function()
        function.reserved_concurrency = 5
        template = self.template_gen.generate([function])
        tf_resource = self.get_function(template)
        assert tf_resource['reserved_concurrent_executions'] == 5

    def test_can_add_tracing_config(self, sample_app):
        function = self.lambda_function()
        function.xray = True
        template = self.template_gen.generate([function])
        tf_resource = self.get_function(template)
        assert tf_resource['tracing_config']['mode'] == 'Active'

    def test_can_generate_cloudwatch_event(self):
        function = self.lambda_function()
        event = models.CloudWatchEvent(
            resource_name='foo-event',
            rule_name='myrule',
            event_pattern='{"source": ["aws.ec2"]}',
            lambda_function=function,
        )
        template = self.template_gen.generate(
            [function, event]
        )
        rule = template['resource'][
            'aws_cloudwatch_event_rule'][event.resource_name]
        assert rule == {
            'name': event.resource_name,
            'event_pattern': event.event_pattern}
        target = template['resource'][
            'aws_cloudwatch_event_target'][event.resource_name]
        assert target == {
            'target_id': 'foo-event',
            'rule': '${aws_cloudwatch_event_rule.foo-event.name}',
            'arn': '${aws_lambda_function.foo.arn}',
        }

    def test_can_generate_scheduled_event(self):
        function = self.lambda_function()
        event = models.ScheduledEvent(
            resource_name='foo-event',
            rule_name='myrule',
            schedule_expression='rate(5 minutes)',
            lambda_function=function,
            rule_description='description',
        )
        template = self.template_gen.generate(
            [function, event]
        )
        rule = template['resource'][
            'aws_cloudwatch_event_rule'][event.resource_name]

        assert rule == {
            'name': event.resource_name,
            'schedule_expression': 'rate(5 minutes)',
            'description': 'description',
        }

    def test_can_generate_rest_api(self, sample_app_with_auth):
        config = Config.create(chalice_app=sample_app_with_auth,
                               project_dir='.',
                               minimum_compression_size=8192,
                               api_gateway_endpoint_type='PRIVATE',
                               api_gateway_endpoint_vpce='vpce-abc123',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        resources = template['resource']
        # Lambda function should be created.
        assert resources['aws_lambda_function']
        # Along with permission to invoke from API Gateway.
        assert list(resources['aws_lambda_permission'].values())[0] == {
            'function_name': 'sample_app-dev',
            'action': 'lambda:InvokeFunction',
            'principal': 'apigateway.amazonaws.com',
            'source_arn': (
                '${aws_api_gateway_rest_api.rest_api.execution_arn}/*')
        }
        assert 'aws_api_gateway_rest_api' in resources
        assert 'rest_api' in resources['aws_api_gateway_rest_api']
        resource_policy = resources[
            'aws_api_gateway_rest_api']['rest_api']['policy']
        assert json.loads(resource_policy) == {
            'Version': '2012-10-17',
            'Statement': [
                {
                    'Action': 'execute-api:Invoke',
                    'Resource': 'arn:*:execute-api:*:*:*',
                    'Effect': 'Allow',
                    'Condition': {
                        'StringEquals': {
                            'aws:SourceVpce': 'vpce-abc123'
                        }
                    },
                    'Principal': '*'
                }
            ]
        }
        assert resources['aws_api_gateway_rest_api'][
            'rest_api']['minimum_compression_size'] == 8192
        assert resources['aws_api_gateway_rest_api'][
            'rest_api']['endpoint_configuration'] == {'types': ['PRIVATE']}

        assert 'aws_api_gateway_stage' not in resources
        assert resources['aws_api_gateway_deployment']['rest_api'] == {
            'rest_api_id': '${aws_api_gateway_rest_api.rest_api.id}',
            'stage_description': (
                '${md5(data.template_file.chalice_api_swagger.rendered)}'),
            'stage_name': 'api',
            'lifecycle': {'create_before_destroy': True}
        }

        # We should also create the auth lambda function.
        assert 'myauth' in resources['aws_lambda_function']

        # Along with permission to invoke from API Gateway.
        assert resources['aws_lambda_permission']['myauth_invoke'] == {
            'action': 'lambda:InvokeFunction',
            'function_name': 'sample_app-dev-myauth',
            'principal': 'apigateway.amazonaws.com',
            'source_arn': (
                '${aws_api_gateway_rest_api.rest_api.execution_arn}/*')
        }

        # Also verify we add the expected outputs when using
        # a Rest API.
        assert template['output'] == {
            'EndpointURL': {
                'value': '${aws_api_gateway_deployment.rest_api.invoke_url}'},
            'RestAPIId': {
                'value': '${aws_api_gateway_rest_api.rest_api.id}'}
        }

    def test_can_package_s3_event_handler_with_tf_ref(self, sample_app):
        @sample_app.on_s3_event(
            bucket='${aws_s3_bucket.my_data_bucket.id}')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        assert template['resource']['aws_s3_bucket_notification'][
                   'my_data_bucket_notify'] == {
                   'bucket': '${aws_s3_bucket.my_data_bucket.id}',
                   'lambda_function': [{
                       'events': ['s3:ObjectCreated:*'],
                       'lambda_function_arn': (
                           '${aws_lambda_function.handler.arn}')
                   }]
               }

    def test_can_generate_chalice_terraform_static_data(self, sample_app):
        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='myfoo',
                               api_gateway_stage='dev')
        template = self.generate_template(config)
        assert template['data']['null_data_source']['chalice']['inputs'] == {
            'app': 'myfoo',
            'stage': 'dev'
        }

    def test_can_package_s3_event_handler_sans_filters(self, sample_app):
        @sample_app.on_s3_event(bucket='foo')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        assert template['resource']['aws_s3_bucket_notification'][
                   'foo_notify'] == {
                   'bucket': 'foo',
                   'lambda_function': [{
                       'events': ['s3:ObjectCreated:*'],
                       'lambda_function_arn': (
                           '${aws_lambda_function.handler.arn}')
                   }]
               }

    def test_can_package_s3_event_handler(self, sample_app):
        @sample_app.on_s3_event(
            bucket='foo', prefix='incoming', suffix='.csv')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        assert template['resource']['aws_lambda_permission'][
                   'handler-s3event'] == {
                   'action': 'lambda:InvokeFunction',
                   'function_name': 'sample_app-dev-handler',
                   'principal': 's3.amazonaws.com',
                   'source_arn': (
                       'arn:${data.aws_partition.chalice.partition}:s3:::foo'),
                   'statement_id': 'handler-s3event'
               }

        assert template['resource']['aws_s3_bucket_notification'][
                   'foo_notify'] == {
                   'bucket': 'foo',
                   'lambda_function': [{
                       'events': ['s3:ObjectCreated:*'],
                       'filter_prefix': 'incoming',
                       'filter_suffix': '.csv',
                       'lambda_function_arn': (
                           '${aws_lambda_function.handler.arn}')
                   }]
               }

    def test_can_package_sns_handler(self, sample_app):
        @sample_app.on_sns_message(topic='foo')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)

        assert template['resource']['aws_sns_topic_subscription'][
                   'handler-sns-subscription'] == {
                   'topic_arn': (
                       'arn:${data.aws_partition.chalice.partition}:sns'
                       ':${data.aws_region.chalice.name}:'
                       '${data.aws_caller_identity.chalice.account_id}:foo'),
                   'protocol': 'lambda',
                   'endpoint': '${aws_lambda_function.handler.arn}'
               }

    def test_can_package_sns_arn_handler(self, sample_app):
        arn = 'arn:aws:sns:space-leo-1:1234567890:foo'

        @sample_app.on_sns_message(topic=arn)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)

        assert template['resource']['aws_sns_topic_subscription'][
                   'handler-sns-subscription'] == {
                   'topic_arn': arn,
                   'protocol': 'lambda',
                   'endpoint': '${aws_lambda_function.handler.arn}'
               }

        assert template['resource']['aws_lambda_permission'][
                   'handler-sns-subscription'] == {
                   'function_name': 'sample_app-dev-handler',
                   'action': 'lambda:InvokeFunction',
                   'principal': 'sns.amazonaws.com',
                   'source_arn': 'arn:aws:sns:space-leo-1:1234567890:foo'
               }

    def test_can_package_sqs_handler(self, sample_app):
        @sample_app.on_sqs_message(queue='foo', batch_size=5)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)

        assert template['resource'][
                   'aws_lambda_event_source_mapping'][
                   'handler-sqs-event-source'] == {
                   'event_source_arn': (
                       'arn:${data.aws_partition.chalice.partition}:sqs'
                       ':${data.aws_region.chalice.name}:'
                       '${data.aws_caller_identity.chalice.account_id}:foo'),
                   'function_name': 'sample_app-dev-handler',
                   'batch_size': 5
               }

    def test_can_package_kinesis_handler(self, sample_app):
        @sample_app.on_kinesis_record(stream='mystream', batch_size=5,
                                      starting_position='TRIM_HORIZON')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)

        assert template['resource'][
                   'aws_lambda_event_source_mapping'][
                   'handler-kinesis-event-source'] == {
                   'event_source_arn': (
                       'arn:${data.aws_partition.chalice.partition}:kinesis'
                       ':${data.aws_region.chalice.name}:'
                       '${data.aws_caller_identity.chalice.account_id}'
                       ':stream/mystream'),
                   'function_name': 'sample_app-dev-handler',
                   'starting_position': 'TRIM_HORIZON',
                   'batch_size': 5
               }

    def test_can_package_dynamodb_handler(self, sample_app):
        @sample_app.on_dynamodb_record(stream_arn='arn:aws:...:stream',
                                       batch_size=5,
                                       starting_position='TRIM_HORIZON')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        template = self.generate_template(config)

        assert template['resource'][
                   'aws_lambda_event_source_mapping'][
                   'handler-dynamodb-event-source'] == {
                       'event_source_arn': 'arn:aws:...:stream',
                       'function_name': 'sample_app-dev-handler',
                       'starting_position': 'TRIM_HORIZON',
                       'batch_size': 5
                   }

    def test_package_websocket_with_error_message(self, sample_websocket_app):
        config = Config.create(chalice_app=sample_websocket_app,
                               project_dir='.',
                               app_name='sample_app',
                               api_gateway_stage='api')
        with pytest.raises(NotImplementedError) as excinfo:
            self.generate_template(config)

        # Should mention the decorator name.
        assert 'Websocket decorators' in str(excinfo.value)
        # Should mention you can use `chalice deploy`.
        assert 'chalice deploy' in str(excinfo.value)

    def test_can_generate_custom_domain_name(self, sample_app):
        config = Config.create(
            chalice_app=sample_app,
            project_dir='.',
            api_gateway_stage='api',
            api_gateway_endpoint_type='EDGE',
            api_gateway_custom_domain={
                "certificate_arn": "my_cert_arn",
                "domain_name": "example.com",
                "tls_version": "TLS_1_2",
                "tags": {"foo": "bar"},
            }
        )
        template = self.generate_template(config)
        assert template['resource']['aws_api_gateway_domain_name'][
            'api_gateway_custom_domain'] == {
                'domain_name': 'example.com',
                'certificate_arn': 'my_cert_arn',
                'security_policy': 'TLS_1_2',
                'endpoint_configuration': {'types': ['EDGE']},
                'tags': {'foo': 'bar'},
        }
        assert template['resource']['aws_api_gateway_base_path_mapping'][
            'api_gateway_custom_domain_mapping'] == {
                'api_id': '${aws_api_gateway_rest_api.rest_api.id}',
                'stage_name': 'api',
                'domain_name': 'example.com',
        }
        outputs = template['output']
        assert outputs['AliasDomainName']['value'] == (
            '${aws_api_gateway_domain_name.api_gateway_custom_domain'
            '.cloudfront_domain_name}'
        )
        assert outputs['HostedZoneId']['value'] == (
            '${aws_api_gateway_domain_name.api_gateway_custom_domain'
            '.cloudfront_zone_id}'
        )

    def test_can_generate_domain_for_regional_endpoint(self, sample_app):
        config = Config.create(
            chalice_app=sample_app,
            project_dir='.',
            api_gateway_stage='api',
            api_gateway_endpoint_type='REGIONAL',
            api_gateway_custom_domain={
                "certificate_arn": "my_cert_arn",
                "domain_name": "example.com",
            }
        )
        template = self.generate_template(config)
        assert template['resource']['aws_api_gateway_domain_name'][
            'api_gateway_custom_domain'] == {
                'domain_name': 'example.com',
                'regional_certificate_arn': 'my_cert_arn',
                'endpoint_configuration': {'types': ['REGIONAL']},
        }
        assert template['resource']['aws_api_gateway_base_path_mapping'][
            'api_gateway_custom_domain_mapping'] == {
                'api_id': '${aws_api_gateway_rest_api.rest_api.id}',
                'stage_name': 'api',
                'domain_name': 'example.com',
        }
        outputs = template['output']
        assert outputs['AliasDomainName']['value'] == (
            '${aws_api_gateway_domain_name.api_gateway_custom_domain'
            '.regional_domain_name}'
        )
        assert outputs['HostedZoneId']['value'] == (
            '${aws_api_gateway_domain_name.api_gateway_custom_domain'
            '.regional_zone_id}'
        )


class TestSAMTemplate(TemplateTestBase):

    template_gen_factory = package.SAMTemplateGenerator

    def test_sam_generates_sam_template_basic(self, sample_app):
        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               automatic_layer=True,
                               api_gateway_stage='api')
        template = self.generate_template(config)
        # Verify the basic structure is in place.  The specific parts
        # are validated in other tests.
        assert template['AWSTemplateFormatVersion'] == '2010-09-09'
        assert template['Transform'] == 'AWS::Serverless-2016-10-31'
        assert 'Outputs' in template
        assert 'Resources' in template
        assert list(sorted(template['Resources'])) == [
            'APIHandler',
            'APIHandlerInvokePermission',
            # This casing on the ApiHandlerRole name is unfortunate, but the 3
            # other resources in this list are hardcoded from the old deployer.
            'ApiHandlerRole',
            'ManagedLayer',
            'RestAPI'
        ]

    def test_can_generate_lambda_layer_if_configured(self, sample_app):
        config = Config.create(chalice_app=sample_app,
                               app_name='testapp',
                               project_dir='.',
                               automatic_layer=True,
                               api_gateway_stage='api')
        template = self.generate_template(config)
        managed_layer = template['Resources']['ManagedLayer']
        assert managed_layer == {
            'Type': 'AWS::Serverless::LayerVersion',
            'Properties': {
                'CompatibleRuntimes': [config.lambda_python_version],
                'LayerName': 'testapp-dev-managed-layer',
                'ContentUri': models.Placeholder.BUILD_STAGE,
            }
        }
        assert template['Resources']['APIHandler']['Properties']['Layers'] == [
            {'Ref': 'ManagedLayer'}
        ]

    def test_adds_single_layer_for_multiple_lambdas(self, sample_app):
        config = Config.create(chalice_app=sample_app,
                               app_name='testapp',
                               project_dir='.',
                               automatic_layer=True,
                               layers=['arn:aws:mylayer'],
                               api_gateway_stage='api')

        @sample_app.lambda_function()
        def first(event, context):
            pass

        @sample_app.lambda_function()
        def second(event, context):
            pass

        template = self.generate_template(config)
        managed_layer = template['Resources']['ManagedLayer']
        assert managed_layer == {
            'Type': 'AWS::Serverless::LayerVersion',
            'Properties': {
                'CompatibleRuntimes': [config.lambda_python_version],
                'LayerName': 'testapp-dev-managed-layer',
                'ContentUri': models.Placeholder.BUILD_STAGE,
            }
        }
        assert template['Resources']['First']['Properties']['Layers'] == [
            {'Ref': 'ManagedLayer'},
            'arn:aws:mylayer',
        ]

    def test_supports_precreated_role(self):
        builder = DependencyBuilder()
        resources = builder.build_dependencies(
            models.Application(
                stage='dev',
                resources=[self.lambda_function()],
            )
        )
        template = self.template_gen.generate(resources)
        assert template['Resources']['Foo']['Properties']['Role'] == 'role:arn'

    def test_sam_injects_policy(self, sample_app):
        function = models.LambdaFunction(
            resource_name='foo',
            function_name='app-dev-foo',
            environment_variables={},
            runtime='python27',
            handler='app.app',
            tags={'foo': 'bar'},
            timeout=120,
            memory_size=128,
            xray=None,
            deployment_package=models.DeploymentPackage(filename='foo.zip'),
            role=models.ManagedIAMRole(
                resource_name='role',
                role_name='app-role',
                trust_policy={},
                policy=models.AutoGenIAMPolicy(document={'iam': 'policy'}),
            ),
            security_group_ids=[],
            subnet_ids=[],
            layers=[],
            reserved_concurrency=None,
        )
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource == {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'CodeUri': 'foo.zip',
                'Handler': 'app.app',
                'MemorySize': 128,
                'Tracing': 'PassThrough',
                'Role': {'Fn::GetAtt': ['Role', 'Arn']},
                'Runtime': 'python27',
                'Tags': {'foo': 'bar'},
                'Timeout': 120
            },
        }

    def test_adds_env_vars_when_provided(self, sample_app):
        function = self.lambda_function()
        function.environment_variables = {'foo': 'bar'}
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource['Properties']['Environment'] == {
            'Variables': {
                'foo': 'bar'
            }
        }

    def test_adds_vpc_config_when_provided(self):
        function = self.lambda_function()
        function.security_group_ids = ['sg1', 'sg2']
        function.subnet_ids = ['sn1', 'sn2']
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource['Properties']['VpcConfig'] == {
            'SecurityGroupIds': ['sg1', 'sg2'],
            'SubnetIds': ['sn1', 'sn2'],
        }

    def test_adds_reserved_concurrency_when_provided(self, sample_app):
        function = self.lambda_function()
        function.reserved_concurrency = 5
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource['Properties']['ReservedConcurrentExecutions'] == 5

    def test_adds_layers_when_provided(self, sample_app):
        function = self.lambda_function()
        function.layers = ['arn:aws:layer1', 'arn:aws:layer2']
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource['Properties']['Layers'] == [
            'arn:aws:layer1',
            'arn:aws:layer2'
            ]

    def test_duplicate_resource_name_raises_error(self):
        one = self.lambda_function()
        two = self.lambda_function()
        one.resource_name = 'foo_bar'
        two.resource_name = 'foo__bar'
        with pytest.raises(package.DuplicateResourceNameError):
            self.template_gen.generate([one, two])

    def test_role_arn_inserted_when_necessary(self):
        function = models.LambdaFunction(
            resource_name='foo',
            function_name='app-dev-foo',
            environment_variables={},
            runtime='python27',
            handler='app.app',
            tags={'foo': 'bar'},
            timeout=120,
            memory_size=128,
            xray=None,
            deployment_package=models.DeploymentPackage(filename='foo.zip'),
            role=models.PreCreatedIAMRole(role_arn='role:arn'),
            security_group_ids=[],
            subnet_ids=[],
            layers=[],
            reserved_concurrency=None,
        )
        template = self.template_gen.generate([function])
        cfn_resource = list(template['Resources'].values())[0]
        assert cfn_resource == {
            'Type': 'AWS::Serverless::Function',
            'Properties': {
                'CodeUri': 'foo.zip',
                'Handler': 'app.app',
                'MemorySize': 128,
                'Role': 'role:arn',
                'Tracing': 'PassThrough',
                'Runtime': 'python27',
                'Tags': {'foo': 'bar'},
                'Timeout': 120
            },
        }

    def test_can_generate_cloudwatch_event(self):
        function = self.lambda_function()
        event = models.CloudWatchEvent(
            resource_name='foo-event',
            rule_name='myrule',
            event_pattern='{"source": ["aws.ec2"]}',
            lambda_function=function,
        )
        template = self.template_gen.generate(
            [function, event]
        )
        resources = template['Resources']
        assert len(resources) == 1
        cfn_resource = list(resources.values())[0]
        assert cfn_resource['Properties']['Events'] == {
            'FooEvent': {
                'Type': 'CloudWatchEvent',
                'Properties': {
                    'Pattern': {
                        'source': [
                            'aws.ec2'
                        ]
                    }
                },
            },
        }

    def test_can_generate_scheduled_event(self):
        function = self.lambda_function()
        event = models.ScheduledEvent(
            resource_name='foo-event',
            rule_name='myrule',
            rule_description="my rule description",
            schedule_expression='rate(5 minutes)',
            lambda_function=function,
        )
        template = self.template_gen.generate(
            [function, event]
        )
        resources = template['Resources']
        assert len(resources) == 1
        cfn_resource = list(resources.values())[0]
        assert cfn_resource['Properties']['Events'] == {
            'FooEvent': {
                'Type': 'Schedule',
                'Properties': {
                    'Schedule': 'rate(5 minutes)'
                },
            },
        }

    def test_can_generate_rest_api_without_compression(
            self, sample_app_with_auth):
        config = Config.create(chalice_app=sample_app_with_auth,
                               project_dir='.',
                               api_gateway_stage='api',
                               )
        template = self.generate_template(config)
        resources = template['Resources']
        assert 'MinimumCompressionSize' not in \
            resources['RestAPI']['Properties']

    def test_can_generate_rest_api(self, sample_app_with_auth):
        config = Config.create(chalice_app=sample_app_with_auth,
                               project_dir='.',
                               api_gateway_stage='api',
                               minimum_compression_size=100,
                               )
        template = self.generate_template(config)
        resources = template['Resources']
        # Lambda function should be created.
        assert resources['APIHandler']['Type'] == 'AWS::Serverless::Function'
        # Along with permission to invoke from API Gateway.
        assert resources['APIHandlerInvokePermission'] == {
            'Type': 'AWS::Lambda::Permission',
            'Properties': {
                'Action': 'lambda:InvokeFunction',
                'FunctionName': {'Ref': 'APIHandler'},
                'Principal': 'apigateway.amazonaws.com',
                'SourceArn': {
                    'Fn::Sub': [
                        ('arn:${AWS::Partition}:execute-api:${AWS::Region}'
                         ':${AWS::AccountId}:${RestAPIId}/*'),
                        {'RestAPIId': {'Ref': 'RestAPI'}}]}},
        }
        assert resources['RestAPI']['Type'] == 'AWS::Serverless::Api'
        assert resources['RestAPI']['Properties']['MinimumCompressionSize'] \
            == 100
        # We should also create the auth lambda function.
        assert resources['Myauth']['Type'] == 'AWS::Serverless::Function'
        # Along with permission to invoke from API Gateway.
        assert resources['MyauthInvokePermission'] == {
            'Type': 'AWS::Lambda::Permission',
            'Properties': {
                'Action': 'lambda:InvokeFunction',
                'FunctionName': {'Fn::GetAtt': ['Myauth', 'Arn']},
                'Principal': 'apigateway.amazonaws.com',
                'SourceArn': {
                    'Fn::Sub': [
                        ('arn:${AWS::Partition}:execute-api:${AWS::Region}'
                         ':${AWS::AccountId}:${RestAPIId}/*'),
                        {'RestAPIId': {'Ref': 'RestAPI'}}]}},
        }
        # Also verify we add the expected outputs when using
        # a Rest API.
        assert template['Outputs'] == {
            'APIHandlerArn': {
                'Value': {
                    'Fn::GetAtt': ['APIHandler', 'Arn']
                }
            },
            'APIHandlerName': {'Value': {'Ref': 'APIHandler'}},
            'EndpointURL': {
                'Value': {
                    'Fn::Sub': (
                        'https://${RestAPI}.execute-api.'
                        '${AWS::Region}.${AWS::URLSuffix}/api/'
                    )
                }
            },
            'RestAPIId': {'Value': {'Ref': 'RestAPI'}}
        }

    @pytest.mark.parametrize('route_key,route', [
        ('$default', 'WebsocketMessageRoute'),
        ('$connect', 'WebsocketConnectRoute'),
        ('$disconnect', 'WebsocketDisconnectRoute')]
    )
    def test_generate_partial_websocket_api(
            self, route_key, route, sample_websocket_app):
        # Remove all but one websocket route.
        sample_websocket_app.websocket_handlers = {
            name: handler for name, handler in
            sample_websocket_app.websocket_handlers.items()
            if name == route_key
        }
        config = Config.create(chalice_app=sample_websocket_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        resources = template['Resources']

        # Check that the template's deployment only depends on the one route.
        depends_on = resources['WebsocketAPIDeployment'].pop('DependsOn')
        assert [route] == depends_on

    def test_generate_websocket_api(self, sample_websocket_app):
        config = Config.create(chalice_app=sample_websocket_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        resources = template['Resources']

        assert resources['WebsocketAPI']['Type'] == 'AWS::ApiGatewayV2::Api'

        for handler, route in (('WebsocketConnect', '$connect'),
                               ('WebsocketMessage', '$default'),
                               ('WebsocketDisconnect', '$disconnect'),):
            # Lambda function should be created.
            assert resources[handler]['Type'] == 'AWS::Serverless::Function'

            # Along with permission to invoke from API Gateway.
            assert resources['%sInvokePermission' % handler] == {
                'Type': 'AWS::Lambda::Permission',
                'Properties': {
                    'Action': 'lambda:InvokeFunction',
                    'FunctionName': {'Ref': handler},
                    'Principal': 'apigateway.amazonaws.com',
                    'SourceArn': {
                        'Fn::Sub': [
                            (
                                'arn:${AWS::Partition}:execute-api'
                                ':${AWS::Region}:${AWS::'
                                'AccountId}:${WebsocketAPIId}/*'
                            ),
                            {'WebsocketAPIId': {'Ref': 'WebsocketAPI'}}]}},
            }

            # Ensure Integration is created.
            assert resources['%sAPIIntegration' % handler] == {
                'Type': 'AWS::ApiGatewayV2::Integration',
                'Properties': {
                    'ApiId': {
                        'Ref': 'WebsocketAPI'
                    },
                    'ConnectionType': 'INTERNET',
                    'ContentHandlingStrategy': 'CONVERT_TO_TEXT',
                    'IntegrationType': 'AWS_PROXY',
                    'IntegrationUri': {
                        'Fn::Sub': [
                            (
                                'arn:${AWS::Partition}:apigateway'
                                ':${AWS::Region}:lambda:path'
                                '/2015-03-31/functions/arn:${AWS::Partition}'
                                ':lambda:${AWS::Region}:${AWS::AccountId}'
                                ':function:${WebsocketHandler}/invocations'
                            ),
                            {'WebsocketHandler': {'Ref': handler}}
                        ],
                    }
                }
            }

            # Route for the handler.
            assert resources['%sRoute' % handler] == {
                'Type': 'AWS::ApiGatewayV2::Route',
                'Properties': {
                    'ApiId': {
                        'Ref': 'WebsocketAPI'
                    },
                    'RouteKey': route,
                    'Target': {
                        'Fn::Join': [
                            '/',
                            [
                                'integrations',
                                {'Ref': '%sAPIIntegration' % handler},
                            ]
                        ]
                    }
                }
            }

        # Ensure the deployment is created. It must manually depend on
        # the routes since it cannot be created for WebsocketAPI that has no
        # routes. The API has no such implicit contract so CloudFormation can
        # deploy things out of order without the explicit DependsOn.
        depends_on = set(resources['WebsocketAPIDeployment'].pop('DependsOn'))
        assert set(['WebsocketConnectRoute',
                    'WebsocketMessageRoute',
                    'WebsocketDisconnectRoute']) == depends_on
        assert resources['WebsocketAPIDeployment'] == {
            'Type': 'AWS::ApiGatewayV2::Deployment',
            'Properties': {
                'ApiId': {
                    'Ref': 'WebsocketAPI'
                }
            }
        }

        # Ensure the stage is created.
        resources['WebsocketAPIStage'] = {
            'Type': 'AWS::ApiGatewayV2::Stage',
            'Properties': {
                'ApiId': {
                    'Ref': 'WebsocketAPI'
                },
                'DeploymentId': {'Ref': 'WebsocketAPIDeployment'},
                'StageName': 'api',
            }
        }

        # Ensure the outputs are created
        assert template['Outputs'] == {
            'WebsocketConnectHandlerArn': {
                'Value': {
                    'Fn::GetAtt': ['WebsocketConnect', 'Arn']
                }
            },
            'WebsocketConnectHandlerName': {
                'Value': {'Ref': 'WebsocketConnect'}},
            'WebsocketMessageHandlerArn': {
                'Value': {
                    'Fn::GetAtt': ['WebsocketMessage', 'Arn']
                }
            },
            'WebsocketMessageHandlerName': {
                'Value': {'Ref': 'WebsocketMessage'}},
            'WebsocketDisconnectHandlerArn': {
                'Value': {
                    'Fn::GetAtt': ['WebsocketDisconnect', 'Arn']
                }
            },
            'WebsocketDisconnectHandlerName': {'Value': {
                'Ref': 'WebsocketDisconnect'}},
            'WebsocketConnectEndpointURL': {
                'Value': {
                    'Fn::Sub': (
                        'wss://${WebsocketAPI}.execute-api.'
                        '${AWS::Region}.${AWS::URLSuffix}/api/'
                    )
                }
            },
            'WebsocketAPIId': {'Value': {'Ref': 'WebsocketAPI'}}
        }

    def test_managed_iam_role(self):
        role = models.ManagedIAMRole(
            resource_name='default_role',
            role_name='app-dev',
            trust_policy=LAMBDA_TRUST_POLICY,
            policy=models.AutoGenIAMPolicy(document={'iam': 'policy'}),
        )
        template = self.template_gen.generate([role])
        resources = template['Resources']
        assert len(resources) == 1
        cfn_role = resources['DefaultRole']
        assert cfn_role['Type'] == 'AWS::IAM::Role'
        assert cfn_role['Properties']['Policies'] == [
            {'PolicyName': 'DefaultRolePolicy',
             'PolicyDocument': {'iam': 'policy'}}
        ]
        # Verify the trust policy is specific to the region
        assert cfn_role['Properties']['AssumeRolePolicyDocument'] == {
            'Statement': [{'Action': 'sts:AssumeRole',
                           'Effect': 'Allow',
                           'Principal': {'Service': 'lambda.amazonaws.com'},
                           'Sid': ''}],
            'Version': '2012-10-17'}

        # Ensure the RoleName is not in the resource properties
        # so we don't require CAPABILITY_NAMED_IAM.
        assert 'RoleName' not in cfn_role['Properties']

    def test_single_role_generated_for_default_config(self,
                                                      sample_app_lambda_only):
        # The sample_app has one lambda function.
        # We'll add a few more and verify they all share the same role.
        @sample_app_lambda_only.lambda_function()
        def second(event, context):
            pass

        @sample_app_lambda_only.lambda_function()
        def third(event, context):
            pass

        config = Config.create(chalice_app=sample_app_lambda_only,
                               project_dir='.',
                               autogen_policy=True,
                               api_gateway_stage='api')
        template = self.generate_template(config)
        roles = [resource for resource in template['Resources'].values()
                 if resource['Type'] == 'AWS::IAM::Role']
        assert len(roles) == 1
        # The lambda functions should all reference this role.
        functions = [
            resource for resource in template['Resources'].values()
            if resource['Type'] == 'AWS::Serverless::Function'
        ]
        role_names = [
            function['Properties']['Role'] for function in functions
        ]
        assert role_names == [
            {'Fn::GetAtt': ['DefaultRole', 'Arn']},
            {'Fn::GetAtt': ['DefaultRole', 'Arn']},
            {'Fn::GetAtt': ['DefaultRole', 'Arn']},
        ]

    def test_vpc_config_added_to_function(self, sample_app_lambda_only):
        config = Config.create(chalice_app=sample_app_lambda_only,
                               project_dir='.',
                               autogen_policy=True,
                               api_gateway_stage='api',
                               security_group_ids=['sg1', 'sg2'],
                               subnet_ids=['sn1', 'sn2'])
        template = self.generate_template(config)
        resources = template['Resources'].values()
        lambda_fns = [resource for resource in resources
                      if resource['Type'] == 'AWS::Serverless::Function']
        assert len(lambda_fns) == 1

        vpc_config = lambda_fns[0]['Properties']['VpcConfig']
        assert vpc_config['SubnetIds'] == ['sn1', 'sn2']
        assert vpc_config['SecurityGroupIds'] == ['sg1', 'sg2']

    def test_helpful_error_message_on_s3_event(self, sample_app):
        @sample_app.on_s3_event(bucket='foo')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        options = PackageOptions(mock.Mock(spec=TypedAWSClient))
        with pytest.raises(NotImplementedError) as excinfo:
            self.generate_template(config, 'dev', options)
        # Should mention the decorator name.
        assert '@app.on_s3_event' in str(excinfo.value)
        # Should mention you can use `chalice deploy`.
        assert 'chalice deploy' in str(excinfo.value)

    def test_can_package_sns_handler(self, sample_app):
        @sample_app.on_sns_message(topic='foo')
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        sns_handler = template['Resources']['Handler']
        assert sns_handler['Properties']['Events'] == {
            'HandlerSnsSubscription': {
                'Type': 'SNS',
                'Properties': {
                    'Topic': {
                        'Fn::Sub': (
                            'arn:${AWS::Partition}:sns:${AWS::Region}'
                            ':${AWS::AccountId}:foo'
                        )
                    }
                },
            }
        }

    def test_can_package_sns_arn_handler(self, sample_app):
        arn = 'arn:aws:sns:space-leo-1:1234567890:foo'

        @sample_app.on_sns_message(topic=arn)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        sns_handler = template['Resources']['Handler']
        assert sns_handler['Properties']['Events'] == {
            'HandlerSnsSubscription': {
                'Type': 'SNS',
                'Properties': {
                    'Topic': arn,
                }
            }
        }

    def test_can_package_sqs_handler(self, sample_app):
        @sample_app.on_sqs_message(queue='foo', batch_size=5)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        sns_handler = template['Resources']['Handler']
        assert sns_handler['Properties']['Events'] == {
            'HandlerSqsEventSource': {
                'Type': 'SQS',
                'Properties': {
                    'Queue': {
                        'Fn::Sub': (
                            'arn:${AWS::Partition}:sqs:${AWS::Region}'
                            ':${AWS::AccountId}:foo'
                        )
                    },
                    'BatchSize': 5,
                },
            }
        }

    def test_can_package_kinesis_handler(self, sample_app):
        @sample_app.on_kinesis_record(stream='mystream', batch_size=5)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        sns_handler = template['Resources']['Handler']
        assert sns_handler['Properties']['Events'] == {
            'HandlerKinesisEventSource': {
                'Type': 'Kinesis',
                'Properties': {
                    'Stream': {
                        'Fn::Sub': (
                            'arn:${AWS::Partition}:kinesis:${AWS::Region}'
                            ':${AWS::AccountId}:stream/mystream'
                        )
                    },
                    'BatchSize': 5,
                    'StartingPosition': 'LATEST',
                },
            }
        }

    def test_can_package_dynamodb_handler(self, sample_app):
        @sample_app.on_dynamodb_record(stream_arn='arn:aws:...:stream',
                                       batch_size=5)
        def handler(event):
            pass

        config = Config.create(chalice_app=sample_app,
                               project_dir='.',
                               api_gateway_stage='api')
        template = self.generate_template(config)
        ddb_handler = template['Resources']['Handler']
        assert ddb_handler['Properties']['Events'] == {
            'HandlerDynamodbEventSource': {
                'Type': 'DynamoDB',
                'Properties': {
                    'Stream': 'arn:aws:...:stream',
                    'BatchSize': 5,
                    'StartingPosition': 'LATEST',
                },
            }
        }

    def test_can_generate_custom_domain_name(self, sample_app):
        config = Config.create(
            chalice_app=sample_app,
            project_dir='.',
            api_gateway_stage='api',
            api_gateway_endpoint_type='EDGE',
            api_gateway_custom_domain={
                "certificate_arn": "my_cert_arn",
                "domain_name": "example.com",
                "tls_version": "TLS_1_2",
                "tags": {"foo": "bar", "bar": "baz"},
            }
        )
        template = self.generate_template(config)
        domain = template['Resources']['ApiGatewayCustomDomain']
        mapping = template['Resources']['ApiGatewayCustomDomainMapping']
        assert domain == {
            'Type': 'AWS::ApiGateway::DomainName',
            'Properties': {
                'CertificateArn': 'my_cert_arn',
                'DomainName': 'example.com',
                'SecurityPolicy': 'TLS_1_2',
                'EndpointConfiguration': {
                    'Types': ['EDGE']
                },
                'Tags': [
                    {'Key': 'bar',
                     'Value': 'baz'},
                    {'Key': 'foo',
                     'Value': 'bar'}
                ]
            }
        }
        assert mapping == {
            'Type': 'AWS::ApiGateway::BasePathMapping',
            'Properties': {
                'DomainName': {'Ref': 'ApiGatewayCustomDomain'},
                'RestApiId': {'Ref': 'RestAPI'},
                'Stage': {'Ref': 'RestAPI.Stage'},
                'BasePath': '(none)',
            }
        }

    def test_can_generate_domain_for_regional_endpoint(self, sample_app):
        config = Config.create(
            chalice_app=sample_app,
            project_dir='.',
            api_gateway_stage='api',
            api_gateway_endpoint_type='REGIONAL',
            api_gateway_custom_domain={
                "certificate_arn": "my_cert_arn",
                "domain_name": "example.com",
            }
        )
        template = self.generate_template(config)
        domain = template['Resources']['ApiGatewayCustomDomain']
        mapping = template['Resources']['ApiGatewayCustomDomainMapping']
        assert domain == {
            'Type': 'AWS::ApiGateway::DomainName',
            'Properties': {
                'RegionalCertificateArn': 'my_cert_arn',
                'DomainName': 'example.com',
                'EndpointConfiguration': {
                    'Types': ['REGIONAL']
                }
            }
        }
        assert mapping == {
            'Type': 'AWS::ApiGateway::BasePathMapping',
            'Properties': {
                'DomainName': {'Ref': 'ApiGatewayCustomDomain'},
                'RestApiId': {'Ref': 'RestAPI'},
                'Stage': {'Ref': 'RestAPI.Stage'},
                'BasePath': '(none)',
            }
        }

    def test_can_generate_domain_for_ws_endpoint(self, sample_websocket_app):
        config = Config.create(
            chalice_app=sample_websocket_app,
            project_dir='.',
            api_gateway_stage='api',
            websocket_api_custom_domain={
                "certificate_arn": "my_cert_arn",
                "domain_name": "example.com",
                'tags': {'foo': 'bar', 'bar': 'baz'},
            }
        )
        template = self.generate_template(config)
        domain = template['Resources']['WebsocketApiCustomDomain']
        mapping = template['Resources']['WebsocketApiCustomDomainMapping']
        assert domain == {
            'Type': 'AWS::ApiGatewayV2::DomainName',
            'Properties': {
                'DomainName': 'example.com',
                'DomainNameConfigurations': [
                    {'CertificateArn': 'my_cert_arn',
                     'EndpointType': 'REGIONAL'},
                ],
                'Tags': {
                    'foo': 'bar',
                    'bar': 'baz'
                },
            }
        }
        assert mapping == {
            'Type': 'AWS::ApiGatewayV2::ApiMapping',
            'Properties': {
                'DomainName': {'Ref': 'WebsocketApiCustomDomain'},
                'ApiId': {'Ref': 'WebsocketAPI'},
                'Stage': {'Ref': 'WebsocketAPIStage'},
                'ApiMappingKey': '(none)',
            }
        }


class TestTemplateDeepMerger(object):
    def test_can_merge_without_changing_identity(self):
        merger = package.TemplateDeepMerger()
        src = {}
        dst = {}

        result = merger.merge(src, dst)
        assert result is not src
        assert result is not dst
        assert src is not dst

    def test_does_not_mutate(self):
        merger = package.TemplateDeepMerger()
        src = {'foo': 'bar'}
        dst = {'baz': 'buz'}

        merger.merge(src, dst)
        assert src == {'foo': 'bar'}
        assert dst == {'baz': 'buz'}

    def test_can_add_element(self):
        merger = package.TemplateDeepMerger()
        src = {'foo': 'bar'}
        dst = {'baz': 'buz'}

        result = merger.merge(src, dst)
        assert result == {
            'foo': 'bar',
            'baz': 'buz',
        }

    def test_can_replace_element(self):
        merger = package.TemplateDeepMerger()
        src = {'foo': 'bar'}
        dst = {'foo': 'buz'}

        result = merger.merge(src, dst)
        assert result == {
            'foo': 'bar',
        }

    def test_can_merge_list(self):
        merger = package.TemplateDeepMerger()
        src = {'foo': [1, 2, 3]}
        dst = {}

        result = merger.merge(src, dst)
        assert result == {
            'foo': [1, 2, 3],
        }

    def test_can_merge_nested_elements(self):
        merger = package.TemplateDeepMerger()
        src = {
            'foo': {
                'bar': 'baz',
            },
        }
        dst = {
            'foo': {
                'qux': 'quack',
            },
        }

        result = merger.merge(src, dst)
        assert result == {
            'foo': {
                'bar': 'baz',
                'qux': 'quack',
            }
        }

    def test_can_merge_nested_list(self):
        merger = package.TemplateDeepMerger()
        src = {
            'foo': {
                'bar': 'baz',
            },
        }
        dst = {
            'foo': {
                'qux': [1, 2, 3, 4],
            },
        }

        result = merger.merge(src, dst)
        assert result == {
            'foo': {
                'bar': 'baz',
                'qux': [1, 2, 3, 4],
            }
        }

    def test_list_elements_are_replaced(self):
        merger = package.TemplateDeepMerger()
        src = {
            'list': [{'foo': 'bar'}],
        }
        dst = {
            'list': [{'foo': 'buz'}],
        }

        result = merger.merge(src, dst)
        assert result == {
            'list': [{'foo': 'bar'}],
        }

    def test_merge_can_change_type(self):
        merger = package.TemplateDeepMerger()
        src = {
            'key': 'foo',
        }
        dst = {
            'key': 1,
        }

        result = merger.merge(src, dst)
        assert result == {
            'key': 'foo'
        }


@pytest.mark.parametrize('filename,is_yaml', [
    ('extras.yaml', True),
    ('extras.YAML', True),
    ('extras.yml', True),
    ('extras.YML', True),
    ('extras.foo.yml', True),
    ('extras', False),
    ('extras.json', False),
    ('extras.yaml.json', False),
    ('foo/bar/extras.yaml', True),
    ('foo/bar/extras.YAML', True),
])
def test_to_cfn_resource_name(filename, is_yaml):
    assert package.YAMLTemplateSerializer.is_yaml_template(filename) == is_yaml


@pytest.mark.parametrize('yaml_contents,expected', [
    ('foo: bar', {'foo': 'bar'}),
    ('foo: !Ref bar', {'foo': {'Ref': 'bar'}}),
    ('foo: !GetAtt Bar.Baz', {'foo': {'Fn::GetAtt': ['Bar', 'Baz']}}),
    ('foo: !FooBar [!Baz YetAnother, "hello"]',
     {'foo': {'Fn::FooBar': [{'Fn::Baz': 'YetAnother'}, 'hello']}}),
    ('foo: !SomeTag {"a": "1"}', {'foo': {'Fn::SomeTag': {'a': '1'}}}),
    ('foo: !GetAtt Foo.Bar.Baz', {'foo': {'Fn::GetAtt': ['Foo', 'Bar.Baz']}}),
    ('foo: !Condition Other', {'foo': {'Condition': 'Other'}}),
    ('foo: !GetAtt ["a", "b"]', {'foo': {'Fn::GetAtt': ['a', 'b']}}),
])
def test_supports_custom_tags(yaml_contents, expected):
    serialize = package.YAMLTemplateSerializer()
    actual = serialize.load_template(yaml_contents)
    assert actual == expected
    # Also verify we can serialize then parse them back to what we originally
    # loaded.  Note that this is not the same thing as round tripping as
    # we convert things like '!Ref foo' over to {'Ref': 'foo'}.
    yaml_str = serialize.serialize_template(actual).strip()
    reparsed = serialize.load_template(yaml_str)
    assert actual == reparsed



import os
import zipfile
import json
import mock
import hashlib

from pytest import fixture
import pytest

import chalice.deploy.deployer
import chalice.deploy.packager
from chalice.awsclient import TypedAWSClient
import chalice.utils
from chalice.config import Config
from chalice import Chalice
from chalice.deploy.packager import MissingDependencyError
from chalice.deploy.packager import EmptyPackageError
from chalice.deploy.packager import LambdaDeploymentPackager
from chalice.deploy.packager import DependencyBuilder
from chalice.deploy.packager import Package


slow = pytest.mark.slow


@fixture
def chalice_deployer():
    ui = chalice.utils.UI()
    osutils = chalice.utils.OSUtils()
    dependency_builder = mock.Mock(spec=DependencyBuilder)
    d = chalice.deploy.packager.LambdaDeploymentPackager(
        osutils=osutils, dependency_builder=dependency_builder,
        ui=ui
    )
    return d


@fixture
def app_only_packager():
    ui = chalice.utils.UI()
    osutils = chalice.utils.OSUtils()
    dependency_builder = mock.Mock(spec=DependencyBuilder)
    d = chalice.deploy.packager.AppOnlyDeploymentPackager(
        osutils=osutils, dependency_builder=dependency_builder,
        ui=ui
    )
    return d, dependency_builder


@fixture
def layer_packager():
    ui = chalice.utils.UI()
    osutils = chalice.utils.OSUtils()
    dependency_builder = mock.Mock(spec=DependencyBuilder)
    d = chalice.deploy.packager.LayerDeploymentPackager(
        osutils=osutils, dependency_builder=dependency_builder,
        ui=ui
    )
    return d, dependency_builder


def _create_app_structure(tmpdir):
    appdir = tmpdir.mkdir('app')
    appdir.join('app.py').write('# Test app')
    appdir.mkdir('.chalice')
    return appdir


@slow
def test_can_create_deployment_package(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write('# Test app')
    chalice_dir = appdir.join('.chalice')
    chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    # There should now be a zip file created.
    contents = chalice_dir.join('deployments').listdir()
    assert len(contents) == 1
    assert str(contents[0]).endswith('.zip')


@slow
def test_can_inject_latest_app(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write('# Test app v1')
    chalice_dir = appdir.join('.chalice')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')

    # Now suppose we update our app code but not any deps.
    # We can use inject_latest_app.
    appdir.join('app.py').write('# Test app NEW VERSION')
    # There should now be a zip file created.
    chalice_deployer.inject_latest_app(name, str(appdir))
    contents = chalice_dir.join('deployments').listdir()
    assert len(contents) == 1
    assert str(contents[0]) == name
    with zipfile.ZipFile(name) as f:
        contents = f.read('app.py')
        assert contents == b'# Test app NEW VERSION'


@slow
def test_zipfile_hash_only_based_on_contents(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write('# Test app v1')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with open(name, 'rb') as f:
        original_checksum = hashlib.md5(f.read()).hexdigest()

    # Now we'll modify the file our app file with the same contents
    # but it will change the mtime.
    app_file = appdir.join('app.py')
    app_file.write('# Test app v1')
    # Set the mtime to something different (1990-1-1T00:00:00).
    # This would normally result in the zipfile having a different
    # checksum.
    os.utime(str(app_file), (631152000.0, 631152000.0))
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with open(name, 'rb') as f:
        new_checksum = hashlib.md5(f.read()).hexdigest()
    assert new_checksum == original_checksum


@slow
def test_app_injection_still_compresses_file(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write('# Test app v1')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    original_size = os.path.getsize(name)
    appdir.join('app.py').write('# Test app v2')
    chalice_deployer.inject_latest_app(name, str(appdir))
    new_size = os.path.getsize(name)
    # The new_size won't be exactly the same as the original,
    # we just want to make sure it wasn't converted to
    # ZIP_STORED, so there's a 5% tolerance.
    assert new_size < (original_size * 1.05)


@slow
def test_no_error_message_printed_on_empty_reqs_file(tmpdir,
                                                     chalice_deployer,
                                                     capfd):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write('# Foo')
    appdir.join('requirements.txt').write('\n')
    chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    out, err = capfd.readouterr()
    assert err.strip() == ''


def test_osutils_proxies_os_functions(tmpdir):
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write(b'hello')

    osutils = chalice.utils.OSUtils()

    app_file = str(appdir.join('app.py'))
    assert osutils.file_exists(app_file)
    assert osutils.get_file_contents(app_file) == b'hello'
    assert osutils.open(app_file, 'rb').read() == b'hello'
    osutils.remove_file(app_file)
    # Removing again doesn't raise an error.
    osutils.remove_file(app_file)
    assert not osutils.file_exists(app_file)


@slow
def test_includes_app_and_chalicelib_dir(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    # We're now also going to create additional files
    chalicelib = appdir.mkdir('chalicelib')
    appdir.join('chalicelib', '__init__.py').write('# Test package')
    appdir.join('chalicelib', 'mymodule.py').write('# Test module')
    appdir.join('chalicelib', 'config.json').write('{"test": "config"}')
    # Should also include sub directories
    subdir = chalicelib.mkdir('subdir')
    subdir.join('submodule.py').write('# Test submodule')
    subdir.join('subconfig.json').write('{"test": "subconfig"}')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('chalicelib/__init__.py', b'# Test package', f)
        _assert_in_zip('chalicelib/mymodule.py', b'# Test module', f)
        _assert_in_zip('chalicelib/config.json', b'{"test": "config"}', f)
        _assert_in_zip('chalicelib/subdir/submodule.py',
                       b'# Test submodule', f)
        _assert_in_zip('chalicelib/subdir/subconfig.json',
                       b'{"test": "subconfig"}', f)


def _assert_in_zip(path, contents, zip):
    allfiles = zip.namelist()
    assert path in allfiles
    assert zip.read(path) == contents


def _assert_not_in_zip(path, zip):
    allfiles = zip.namelist()
    assert path not in allfiles


@slow
def test_subsequent_deploy_replaces_chalicelib(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    chalicelib = appdir.mkdir('chalicelib')
    appdir.join('chalicelib', '__init__.py').write('# Test package')
    subdir = chalicelib.mkdir('subdir')
    subdir.join('submodule.py').write('# Test submodule')

    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    subdir.join('submodule.py').write('# Test submodule v2')
    appdir.join('chalicelib', '__init__.py').remove()
    chalice_deployer.inject_latest_app(name, str(appdir))
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('chalicelib/subdir/submodule.py',
                       b'# Test submodule v2', f)
        # And chalicelib/__init__.py should no longer be
        # in the zipfile because we deleted it in the appdir.
        assert 'chalicelib/__init__.py' not in f.namelist()


@slow
def test_vendor_dir_included(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    vendor = appdir.mkdir('vendor')
    extra_package = vendor.mkdir('mypackage')
    extra_package.join('__init__.py').write('# Test package')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('mypackage/__init__.py', b'# Test package', f)


@slow
def test_no_vendor_in_app_only_packager(tmpdir, app_only_packager):
    packager, deps_builder = app_only_packager
    appdir = _create_app_structure(tmpdir)
    appdir.mkdir('chalicelib')
    appdir.join('requirements.txt').write('boto3')
    appdir.join('chalicelib', '__init__.py').write('# Test package')
    vendor = appdir.mkdir('vendor')
    extra_package = vendor.mkdir('mypackage')
    extra_package.join('__init__.py').write('# Test package')
    name = packager.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_not_in_zip('mypackage/__init__.py', f)
        _assert_in_zip('chalicelib/__init__.py', b'# Test package', f)
        _assert_in_zip('app.py', b'# Test app', f)
    assert not deps_builder.build_site_packages.called


@slow
def test_py_deps_in_layer_package(tmpdir, layer_packager):
    packager, deps_builder = layer_packager
    appdir = _create_app_structure(tmpdir)
    appdir.mkdir('chalicelib')
    appdir.join('requirements.txt').write('boto3')
    appdir.join('chalicelib', '__init__.py').write('# Test package')
    vendor = appdir.mkdir('vendor')
    extra_package = vendor.mkdir('mypackage')
    extra_package.join('__init__.py').write('# Test package')
    name = packager.create_deployment_package(
        str(appdir), 'python2.7')
    assert os.path.basename(name).startswith('managed-layer-')
    with zipfile.ZipFile(name) as f:
        prefix = 'python/lib/python2.7/site-packages'
        _assert_in_zip(
            '%s/mypackage/__init__.py' % prefix, b'# Test package', f)
        _assert_not_in_zip('%s/chalicelib/__init__.py' % prefix, f)
        _assert_not_in_zip('%s/app.py' % prefix, f)
    deps_builder.build_site_packages.assert_called_with(
        'cp27mu', str(appdir.join('requirements.txt')), mock.ANY
    )


def test_empty_layer_package_raises_error(tmpdir, layer_packager):
    packager, deps_builder = layer_packager
    appdir = _create_app_structure(tmpdir)
    appdir.mkdir('chalicelib')
    appdir.join('requirements.txt').write('')
    appdir.join('chalicelib', '__init__.py').write('# Test package')
    filename = packager.deployment_package_filename(str(appdir), 'python2.7')
    with pytest.raises(EmptyPackageError):
        packager.create_deployment_package(
            str(appdir), 'python2.7')
    # We should also verify that the file does not exist so it doesn't
    # get reused in subsequent caches.  This shouldn't affect anything,
    # we're just trying to cleanup properly.
    assert not os.path.isfile(filename)


@slow
def test_subsequent_deploy_replaces_vendor_dir(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    vendor = appdir.mkdir('vendor')
    extra_package = vendor.mkdir('mypackage')
    extra_package.join('__init__.py').write('# v1')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    # Now we update a package in vendor/ with a new version.
    extra_package.join('__init__.py').write('# v2')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('mypackage/__init__.py', b'# v2', f)


@slow
def test_vendor_symlink_included(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    extra_package = tmpdir.mkdir('mypackage')
    extra_package.join('__init__.py').write('# Test package')
    vendor = appdir.mkdir('vendor')
    os.symlink(str(extra_package), str(vendor.join('otherpackage')))
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('otherpackage/__init__.py', b'# Test package', f)


@slow
def test_subsequent_deploy_replaces_vendor_symlink(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    extra_package = tmpdir.mkdir('mypackage')
    extra_package.join('__init__.py').write('# v1')
    vendor = appdir.mkdir('vendor')
    os.symlink(str(extra_package), str(vendor.join('otherpackage')))
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('otherpackage/__init__.py', b'# v1', f)
    # Now we update a package in vendor/ with a new version.
    extra_package.join('__init__.py').write('# v2')
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    with zipfile.ZipFile(name) as f:
        _assert_in_zip('otherpackage/__init__.py', b'# v2', f)


def test_zip_filename_changes_on_vendor_update(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    vendor = appdir.mkdir('vendor')
    extra_package = vendor.mkdir('mypackage')
    extra_package.join('__init__.py').write('# v1')
    first = chalice_deployer.deployment_package_filename(
        str(appdir), 'python3.6')
    extra_package.join('__init__.py').write('# v2')
    second = chalice_deployer.deployment_package_filename(
        str(appdir), 'python3.6')
    assert first != second


def test_zip_filename_changes_on_vendor_symlink(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    vendor = appdir.mkdir('vendor')
    extra_package = tmpdir.mkdir('mypackage')
    extra_package.join('__init__.py').write('# v1')
    os.symlink(str(extra_package), str(vendor.join('otherpackage')))
    first = chalice_deployer.deployment_package_filename(
        str(appdir), 'python3.6')
    extra_package.join('__init__.py').write('# v2')
    second = chalice_deployer.deployment_package_filename(
        str(appdir), 'python3.6')
    assert first != second


@slow
def test_chalice_runtime_injected_on_change(tmpdir, chalice_deployer):
    appdir = _create_app_structure(tmpdir)
    name = chalice_deployer.create_deployment_package(
        str(appdir), 'python2.7')
    # We're verifying that we always inject the chalice runtime
    # but we can't actually modify the runtime in this repo, so
    # instead we'll modify the deployment package and change the
    # runtime.
    # We'll then verify when we inject the latest app the runtime
    # has been re-added.  This should give us enough confidence
    # that the runtime is always being inserted.
    _remove_runtime_from_deployment_package(name)
    with zipfile.ZipFile(name) as z:
        assert 'chalice/app.py' not in z.namelist()
    chalice_deployer.inject_latest_app(name, str(appdir))
    with zipfile.ZipFile(name) as z:
        assert 'chalice/app.py' in z.namelist()


def test_does_handle_missing_dependency_error(tmpdir):
    appdir = _create_app_structure(tmpdir)
    builder = mock.Mock(spec=DependencyBuilder)
    fake_package = mock.Mock(spec=Package)
    fake_package.identifier = 'foo==1.2'
    builder.build_site_packages.side_effect = MissingDependencyError(
        set([fake_package]))
    ui = mock.Mock(spec=chalice.utils.UI)
    osutils = chalice.utils.OSUtils()
    packager = LambdaDeploymentPackager(
        osutils=osutils,
        dependency_builder=builder,
        ui=ui,
    )
    packager.create_deployment_package(str(appdir), 'python2.7')

    output = ''.join([call[0][0] for call in ui.write.call_args_list])
    assert 'Could not install dependencies:\nfoo==1.2' in output


def _remove_runtime_from_deployment_package(filename):
    new_filename = os.path.join(os.path.dirname(filename), 'new.zip')
    with zipfile.ZipFile(filename, 'r') as original:
        with zipfile.ZipFile(new_filename, 'w',
                             compression=zipfile.ZIP_DEFLATED) as z:
            for item in original.infolist():
                if item.filename.startswith('chalice/'):
                    continue
                contents = original.read(item.filename)
                z.writestr(item, contents)
    os.remove(filename)
    os.rename(new_filename, filename)


def test_can_delete_app(tmpdir):
    # This is just a sanity check that deletions are working
    # as expected now that there's no separate interface
    # for deletion at the deployer layer.
    appdir = _create_app_structure(tmpdir)
    appdir.join('app.py').write(
        'from chalice import Chalice\n'
        'app = Chalice("testapp")'
    )
    deployed_json = {
        'schema_version': '2.0',
        'backend': 'api',
        'resources': [
            {'name': 'role-index',
                'resource_type': 'iam_role',
                'role_name': 'testapp-dev-index',
                'role_arn': 'arn:aws:iam::1:role/testapp-dev-index'},
            {'lambda_arn': 'arn:aws:lambda:r:1:f:testapp-dev-index',
                'name': 'index', 'resource_type': 'lambda_function'},
            {'name': 'role-james', 'resource_type': 'iam_role',
                'role_name': 'testapp-dev-foo',
                'role_arn': 'arn:aws:iam::1:role/testapp-dev-foo'},
            {'lambda_arn': 'arn:aws:lambda:r:1:f:testapp-dev-foo',
                'name': 'james', 'resource_type': 'lambda_function'}
        ]
    }
    deployed_dir = appdir.join('.chalice', 'deployed')
    deployed_dir.mkdir()
    deployed_dir.join('dev.json').write(
        json.dumps(deployed_json))
    mock_client = mock.Mock(spec=TypedAWSClient)
    ui = mock.Mock(spec=chalice.utils.UI)
    d = chalice.deploy.deployer.create_deletion_deployer(mock_client, ui)

    config = Config(
        chalice_stage='dev',
        user_provided_params={
            'chalice_app': Chalice('testapp'),
            'project_dir': str(appdir),
        },
        config_from_disk={},
        default_params={}
    )
    returned_values = d.deploy(config, 'dev')
    assert returned_values == {
        'schema_version': '2.0',
        'backend': 'api',
        'resources': [],
    }
    call = mock.call
    expected_calls = [
        call.delete_function(
            function_name=u'arn:aws:lambda:r:1:f:testapp-dev-foo'),
        call.delete_role(name=u'testapp-dev-foo'),
        call.delete_function(
            function_name=u'arn:aws:lambda:r:1:f:testapp-dev-index'),
        call.delete_role(name=u'testapp-dev-index')
    ]
    assert expected_calls == mock_client.method_calls



import pytest
import mock
import threading
import os
import unittest
import time

from chalice.cli.filewatch.stat import StatWorkerProcess
try:
    from chalice.cli.filewatch.eventbased import WatchdogWorkerProcess
    WATCHDOG_AVAILABLE = True
except ImportError:
    WATCHDOG_AVAILABLE = False

import chalice.local


DEFAULT_DELAY = 0.1
SETTLE_DELAY = 1
MAX_TIMEOUT = 5.0
use_all_watcher_types = pytest.mark.parametrize(
    ['worker_class_type'], [('watchdog',), ('stat',)])


def modify_file_after_n_seconds(filename, contents, delay=DEFAULT_DELAY):
    t = threading.Timer(delay, function=modify_file, args=(filename, contents))
    t.daemon = True
    t.start()


def delete_file_after_n_seconds(filename, delay=DEFAULT_DELAY):
    t = threading.Timer(delay, function=os.remove, args=(filename,))
    t.daemon = True
    t.start()


def modify_file(filename, contents):
    if filename is None:
        return
    with open(filename, 'w') as f:
        f.write(contents)


def assert_reload_happens(root_dir, when_modified_file, using_worker_class):
    http_thread = mock.Mock(spec=chalice.local.HTTPServerThread)
    worker_cls = get_worker_cls(using_worker_class)
    p = worker_cls(http_thread)
    if isinstance(when_modified_file, tuple):
        if when_modified_file[1] == 'is_deleted':
            delete_file_after_n_seconds(when_modified_file[0])
    else:
        modify_file_after_n_seconds(when_modified_file, 'contents')
    rc = p.main(root_dir, MAX_TIMEOUT)
    assert rc == chalice.cli.filewatch.RESTART_REQUEST_RC


def get_worker_cls(worker_class_name):
    if worker_class_name == 'watchdog':
        if not WATCHDOG_AVAILABLE:
            raise unittest.SkipTest("Test requires watchdog package.")
        else:
            return WatchdogWorkerProcess
    elif worker_class_name == 'stat':
        return StatWorkerProcess
    else:
        raise RuntimeError("Unknown worker class type name: %s"
                           % worker_class_name)


@use_all_watcher_types
def test_can_reload_when_file_created(tmpdir, worker_class_type):
    top_level_file = str(tmpdir.join('foo'))
    assert_reload_happens(str(tmpdir), when_modified_file=top_level_file,
                          using_worker_class=worker_class_type)


@use_all_watcher_types
def test_can_reload_when_subdir_file_created(tmpdir, worker_class_type):
    subdir_file = str(tmpdir.join('subdir').mkdir().join('foo.txt'))
    assert_reload_happens(str(tmpdir), when_modified_file=subdir_file,
                          using_worker_class=worker_class_type)


@use_all_watcher_types
def test_can_reload_when_file_modified(tmpdir, worker_class_type):
    top_level_file = tmpdir.join('foo')
    top_level_file.write('original contents')
    # If you write to the file and immediately start the reloader, it
    # won't see the initial write() above.  I tried out a few delay options,
    # and a separate SETTLE_DELAY was necessary in order to prevent
    # intermittent failures.
    time.sleep(SETTLE_DELAY)
    assert_reload_happens(str(tmpdir), when_modified_file=str(top_level_file),
                          using_worker_class=worker_class_type)


@use_all_watcher_types
def test_can_reload_when_file_removed(tmpdir, worker_class_type):
    top_level_file = tmpdir.join('foo')
    top_level_file.write('original contents')
    assert_reload_happens(
        str(tmpdir), when_modified_file=(str(top_level_file), 'is_deleted'),
        using_worker_class=worker_class_type
    )


@use_all_watcher_types
def test_rc_0_when_no_file_modified(tmpdir, worker_class_type):
    http_thread = mock.Mock(spec=chalice.local.HTTPServerThread)
    worker_cls = get_worker_cls(worker_class_type)
    p = worker_cls(http_thread)
    rc = p.main(str(tmpdir), timeout=0.2)
    assert rc == 0



import json

import pytest

from chalice.awsclient import TypedAWSClient
from chalice.invoke import LambdaInvoker
from chalice.invoke import LambdaInvokeHandler
from chalice.invoke import LambdaResponseFormatter
from chalice.invoke import UnhandledLambdaError


class FakeUI(object):
    def __init__(self):
        self.writes = []
        self.errors = []

    def write(self, value):
        self.writes.append(value)

    def error(self, value):
        self.errors.append(value)


class FakeStreamingBody(object):
    def __init__(self, value):
        self._value = value

    def read(self):
        return self._value


class TestLambdaInvokeHandler(object):
    def test_invoke_can_format_and_write_success_case(self, stubbed_session):
        arn = 'arn:aws:lambda:region:id:function:name-dev'
        stubbed_session.stub('lambda').invoke(
            FunctionName=arn,
            InvocationType='RequestResponse'
        ).returns({
            'StatusCode': 200,
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(b'foobarbaz')
        })
        ui = FakeUI()
        stubbed_session.activate_stubs()
        client = TypedAWSClient(stubbed_session)
        invoker = LambdaInvoker(arn, client)
        formatter = LambdaResponseFormatter()
        invoke_handler = LambdaInvokeHandler(invoker, formatter, ui)
        invoke_handler.invoke()

        stubbed_session.verify_stubs()
        assert ['foobarbaz\n'] == ui.writes

    def test_invoke_can_format_and_write_error_case(self, stubbed_session):
        arn = 'arn:aws:lambda:region:id:function:name-dev'
        error = {
            "errorMessage": "Something bad happened",
            "errorType": "Error",
            "stackTrace": [
                ["/path/file.py", 123, "main", "foo(bar)"],
                ["/path/other_file.py", 456, "function", "bar = baz"]
            ]
        }
        serialized_error = json.dumps(error).encode('utf-8')
        stubbed_session.stub('lambda').invoke(
            FunctionName=arn,
            InvocationType='RequestResponse'
        ).returns({
            'StatusCode': 200,
            'FunctionError': 'Unhandled',
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(serialized_error)
        })
        ui = FakeUI()
        stubbed_session.activate_stubs()
        client = TypedAWSClient(stubbed_session)
        invoker = LambdaInvoker(arn, client)
        formatter = LambdaResponseFormatter()
        invoke_handler = LambdaInvokeHandler(invoker, formatter, ui)
        with pytest.raises(UnhandledLambdaError):
            invoke_handler.invoke()

        stubbed_session.verify_stubs()
        assert [(
            'Traceback (most recent call last):\n'
            '  File "/path/file.py", line 123, in main\n'
            '    foo(bar)\n'
            '  File "/path/other_file.py", line 456, in function\n'
            '    bar = baz\n'
            'Error: Something bad happened\n'
        )] == ui.errors

    def test_invoke_can_format_and_write_small_error_case(self,
                                                          stubbed_session):
        # Some error response payloads do not have the errorType or
        # stackTrace key.
        arn = 'arn:aws:lambda:region:id:function:name-dev'
        error = {
            "errorMessage": "Something bad happened",
        }
        serialized_error = json.dumps(error).encode('utf-8')
        stubbed_session.stub('lambda').invoke(
            FunctionName=arn,
            InvocationType='RequestResponse'
        ).returns({
            'StatusCode': 200,
            'FunctionError': 'Unhandled',
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(serialized_error)
        })
        ui = FakeUI()
        stubbed_session.activate_stubs()
        client = TypedAWSClient(stubbed_session)
        invoker = LambdaInvoker(arn, client)
        formatter = LambdaResponseFormatter()
        invoke_handler = LambdaInvokeHandler(invoker, formatter, ui)
        with pytest.raises(UnhandledLambdaError):
            invoke_handler.invoke()

        stubbed_session.verify_stubs()
        assert [(
            'Something bad happened\n'
        )] == ui.errors


class TestLambdaInvoker(object):
    def test_invoke_can_call_api_handler(self, stubbed_session):
        arn = 'arn:aws:lambda:region:id:function:name-dev'
        stubbed_session.stub('lambda').invoke(
            FunctionName=arn,
            InvocationType='RequestResponse'
        ).returns({})
        stubbed_session.activate_stubs()
        client = TypedAWSClient(stubbed_session)
        invoker = LambdaInvoker(arn, client)
        invoker.invoke()
        stubbed_session.verify_stubs()

    def test_invoke_does_forward_payload(self, stubbed_session):
        arn = 'arn:aws:lambda:region:id:function:name-dev'
        stubbed_session.stub('lambda').invoke(
            FunctionName=arn,
            InvocationType='RequestResponse',
            Payload=b'foobar',
        ).returns({})
        stubbed_session.activate_stubs()
        client = TypedAWSClient(stubbed_session)
        invoker = LambdaInvoker(arn, client)
        invoker.invoke(b'foobar')
        stubbed_session.verify_stubs()


class TestLambdaResponseFormatter(object):
    def test_formatter_can_format_success(self):
        formatter = LambdaResponseFormatter()
        formatted = formatter.format_response({
            'StatusCode': 200,
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(b'foobarbaz')
        })
        assert 'foobarbaz\n' == formatted

    def test_formatter_can_format_list_stack_trace(self):
        error = {
            "errorMessage": "Something bad happened",
            "errorType": "Error",
            "stackTrace": [
                ["/path/file.py", 123, "main", "foo(bar)"],
                ["/path/more.py", 456, "func", "bar = baz"]
            ]
        }
        serialized_error = json.dumps(error).encode('utf-8')
        formatter = LambdaResponseFormatter()
        formatted = formatter.format_response({
            'StatusCode': 200,
            'FunctionError': 'Unhandled',
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(serialized_error)
        })

        assert (
            'Traceback (most recent call last):\n'
            '  File "/path/file.py", line 123, in main\n'
            '    foo(bar)\n'
            '  File "/path/more.py", line 456, in func\n'
            '    bar = baz\n'
            'Error: Something bad happened\n'
        ) == formatted

    def test_formatter_can_format_string_stack_trace(self):
        error = {
            "errorMessage": "Something bad happened",
            "errorType": "Error",
            "stackTrace": [
                '  File "/path/file.py", line 123, in main\n    foo(bar)\n',
                '  File "/path/more.py", line 456, in func\n    bar = baz\n',
            ]
        }
        serialized_error = json.dumps(error).encode('utf-8')
        formatter = LambdaResponseFormatter()
        formatted = formatter.format_response({
            'StatusCode': 200,
            'FunctionError': 'Unhandled',
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(serialized_error)
        })

        assert (
            'Traceback (most recent call last):\n'
            '  File "/path/file.py", line 123, in main\n'
            '    foo(bar)\n'
            '  File "/path/more.py", line 456, in func\n'
            '    bar = baz\n'
            'Error: Something bad happened\n'
        ) == formatted

    def test_formatter_can_format_simple_error(self):
        error = {
            "errorMessage": "Something bad happened",
        }
        serialized_error = json.dumps(error).encode('utf-8')
        formatter = LambdaResponseFormatter()
        formatted = formatter.format_response({
            'StatusCode': 200,
            'FunctionError': 'Unhandled',
            'ExecutedVersion': '$LATEST',
            'Payload': FakeStreamingBody(serialized_error)
        })

        assert 'Something bad happened\n' == formatted



import json
import zipfile
import os
import sys
import re

import pytest
from click.testing import CliRunner
import mock
from botocore.exceptions import ClientError

from chalice import cli
from chalice.cli import factory
from chalice.config import Config, DeployedResources
from chalice.utils import record_deployed_values
from chalice.utils import PipeReader
from chalice.constants import DEFAULT_APIGATEWAY_STAGE_NAME
from chalice.logs import LogRetriever, LogRetrieveOptions
from chalice.invoke import LambdaInvokeHandler
from chalice.invoke import UnhandledLambdaError
from chalice.awsclient import ReadTimeout
from chalice.deploy.validate import ExperimentalFeatureError


class FakeConfig(object):
    def __init__(self, deployed_resources):
        self._deployed_resources = deployed_resources

    def deployed_resources(self, chalice_stage_name):
        return self._deployed_resources


@pytest.fixture
def runner():
    return CliRunner()


@pytest.fixture
def mock_cli_factory():
    cli_factory = mock.Mock(spec=factory.CLIFactory)
    cli_factory.create_config_obj.return_value = Config.create(project_dir='.')
    cli_factory.create_botocore_session.return_value = mock.sentinel.Session
    return cli_factory


def teardown_function(function):
    sys.modules.pop('app', None)


def assert_chalice_app_structure_created(dirname):
    app_contents = os.listdir(os.path.join(os.getcwd(), dirname))
    assert 'app.py' in app_contents
    assert 'requirements.txt' in app_contents
    assert '.chalice' in app_contents
    assert '.gitignore' in app_contents


def _run_cli_command(runner, function, args, cli_factory=None):
    # Handles passing in 'obj' so we can get commands
    # that use @pass_context to work properly.
    # click doesn't support this natively so we have to duplicate
    # what 'def cli(...)' is doing.
    if cli_factory is None:
        cli_factory = factory.CLIFactory('.')
    result = runner.invoke(
        function, args, obj={'project_dir': '.', 'debug': False,
                             'factory': cli_factory})
    return result


def test_create_new_project_creates_app(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.new_project, ['testproject'])
        assert result.exit_code == 0

        # The 'new-project' command creates a directory based on
        # the project name
        assert os.listdir(os.getcwd()) == ['testproject']
        assert_chalice_app_structure_created(dirname='testproject')


def test_create_project_with_prompted_app_name(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.new_project, input='testproject')
        assert result.exit_code == 0
        assert os.listdir(os.getcwd()) == ['testproject']
        assert_chalice_app_structure_created(dirname='testproject')


def test_error_raised_if_dir_already_exists(runner):
    with runner.isolated_filesystem():
        os.mkdir('testproject')
        result = runner.invoke(cli.new_project, ['testproject'])
        assert result.exit_code == 1
        assert 'Directory already exists: testproject' in result.output


def test_can_load_project_config_after_project_creation(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.new_project, ['testproject'])
        assert result.exit_code == 0
        config = factory.CLIFactory('testproject').load_project_config()
        assert config == {
            'version': '2.0',
            'app_name': 'testproject',
            'stages': {
                'dev': {'api_gateway_stage': DEFAULT_APIGATEWAY_STAGE_NAME},
            }
        }


def test_default_new_project_adds_index_route(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.new_project, ['testproject'])
        assert result.exit_code == 0
        app = factory.CLIFactory('testproject').load_chalice_app()
        assert '/' in app.routes


def test_gen_policy_command_creates_policy(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = runner.invoke(cli.cli, ['gen-policy'], obj={})
        assert result.exit_code == 0
        # The output should be valid JSON.
        parsed_policy = json.loads(result.output)
        # We don't want to validate the specific parts of the policy
        # (that's tested elsewhere), but we'll check to make sure
        # it looks like a policy document.
        assert 'Version' in parsed_policy
        assert 'Statement' in parsed_policy


def test_does_fail_to_generate_swagger_if_no_rest_api(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        with open('app.py', 'w') as f:
            f.write(
                'from chalice import Chalice\n'
                'app = Chalice("myapp")\n'
            )
        result = _run_cli_command(runner, cli.generate_models, [])
        assert result.exit_code == 1
        assert result.output == (
            'No REST API found to generate model from.\n'
            'Aborted!\n'
        )


def test_can_write_swagger_model(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.generate_models, [])
        assert result.exit_code == 0
        model = json.loads(result.output)
        assert model == {
            "swagger": "2.0",
            "info": {
                "version": "1.0",
                "title": "testproject"
            },
            "schemes": [
                "https"
            ],
            "paths": {
                "/": {
                    "get": {
                        "consumes": [
                            "application/json"
                        ],
                        "produces": [
                            "application/json"
                        ],
                        "responses": {
                            "200": {
                                "description": "200 response",
                                "schema": {
                                    "$ref": "#/definitions/Empty"
                                }
                            }
                        },
                        "x-amazon-apigateway-integration": {
                            "responses": {
                                "default": {
                                    "statusCode": "200"
                                }
                            },
                            "uri": (
                                "arn:{partition}:apigateway:{region_name}"
                                ":lambda:path/2015-03-31/functions/"
                                "{api_handler_lambda_arn}/invocations"
                            ),
                            "passthroughBehavior": "when_no_match",
                            "httpMethod": "POST",
                            "contentHandling": "CONVERT_TO_TEXT",
                            "type": "aws_proxy"
                        }
                    }
                }
            },
            "definitions": {
                "Empty": {
                    "type": "object",
                    "title": "Empty Schema"
                }
            },
            "x-amazon-apigateway-binary-media-types": [
                "application/octet-stream",
                "application/x-tar",
                "application/zip",
                "audio/basic",
                "audio/ogg",
                "audio/mp4",
                "audio/mpeg",
                "audio/wav",
                "audio/webm",
                "image/png",
                "image/jpg",
                "image/jpeg",
                "image/gif",
                "video/ogg",
                "video/mpeg",
                "video/webm"
            ]
        }


def test_can_package_command(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.package, ['outdir'])
        assert result.exit_code == 0, result.output
        assert os.path.isdir('outdir')
        dir_contents = os.listdir('outdir')
        assert 'sam.json' in dir_contents
        assert 'deployment.zip' in dir_contents


def test_can_package_with_yaml_command(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.package,
                                  ['--template-format', 'yaml', 'outdir'])
        assert result.exit_code == 0, result.output
        assert os.path.isdir('outdir')
        dir_contents = os.listdir('outdir')
        assert 'sam.yaml' in dir_contents
        assert 'deployment.zip' in dir_contents


def test_case_insensitive_template_format(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.package,
                                  ['--template-format', 'YAML', 'outdir'])
        assert result.exit_code == 0, result.output
        assert os.path.isdir('outdir')
        assert 'sam.yaml' in os.listdir('outdir')


def test_can_package_with_single_file(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.package, ['--single-file', 'package.zip'])
        assert result.exit_code == 0, result.output
        assert os.path.isfile('package.zip')
        with zipfile.ZipFile('package.zip', 'r') as f:
            assert sorted(f.namelist()) == [
                'deployment.zip', 'sam.json']


def test_package_terraform_err_with_single_file_or_merge(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.package, ['--pkg-format', 'terraform',
                                  '--single-file', 'module'])
        assert result.exit_code == 1, result.output
        assert "Terraform format does not support" in result.output

        result = _run_cli_command(
            runner, cli.package, ['--pkg-format', 'terraform',
                                  '--merge-template', 'foo.json', 'module'])
        assert result.exit_code == 1, result.output
        assert "Terraform format does not support" in result.output


def test_debug_flag_enables_logging(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = runner.invoke(
            cli.cli, ['--debug', 'package', 'outdir'], obj={})
        assert result.exit_code == 0
        assert re.search('[DEBUG].*Creating deployment package',
                         result.output) is not None


def test_does_deploy_with_default_api_gateway_stage_name(runner,
                                                         mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        # This isn't perfect as we're assuming we know how to
        # create the config_obj like the deploy() command does,
        # it should give us more confidence that the api gateway
        # stage defaults are still working.
        cli_factory = factory.CLIFactory('.')
        config = cli_factory.create_config_obj(
            chalice_stage_name='dev',
            autogen_policy=None,
            api_gateway_stage=None
        )
        assert config.api_gateway_stage == DEFAULT_APIGATEWAY_STAGE_NAME


def test_can_specify_api_gateway_stage(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.deploy,
                                  ['--api-gateway-stage', 'notdev'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 0
        mock_cli_factory.create_config_obj.assert_called_with(
            autogen_policy=None, chalice_stage_name='dev',
            api_gateway_stage='notdev'
        )


def test_can_deploy_specify_connection_timeout(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.deploy,
                                  ['--connection-timeout', 100],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 0
        mock_cli_factory.create_botocore_session.assert_called_with(
            connection_timeout=100
        )


def test_can_retrieve_url(runner, mock_cli_factory):
    deployed_values_dev = {
        "schema_version": "2.0",
        "resources": [
            {"rest_api_url": "https://dev-url/",
             "name": "rest_api",
             "resource_type": "rest_api"},
        ]
    }
    deployed_values_prod = {
        "schema_version": "2.0",
        "resources": [
            {"rest_api_url": "https://prod-url/",
             "name": "rest_api",
             "resource_type": "rest_api"},
        ]
    }
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        deployed_dir = os.path.join('.chalice', 'deployed')
        os.makedirs(deployed_dir)
        record_deployed_values(
            deployed_values_dev,
            os.path.join(deployed_dir, 'dev.json')
        )
        record_deployed_values(
            deployed_values_prod,
            os.path.join(deployed_dir, 'prod.json')
        )
        result = _run_cli_command(runner, cli.url, [],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 0
        assert result.output == 'https://dev-url/\n'

        prod_result = _run_cli_command(runner, cli.url, ['--stage', 'prod'],
                                       cli_factory=mock_cli_factory)
        assert prod_result.exit_code == 0
        assert prod_result.output == 'https://prod-url/\n'


def test_error_when_no_deployed_record(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.url, [],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 2
        assert 'not find' in result.output


@pytest.mark.skipif(sys.version_info[:2] == (3, 7),
                    reason="Cannot generate pipeline for python3.7.")
@pytest.mark.skipif(sys.version_info[:2] == (3, 8),
                    reason="Cannot generate pipeline for python3.8.")
def test_can_generate_pipeline_for_all(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.generate_pipeline, ['pipeline.json'])
        assert result.exit_code == 0, result.output
        assert os.path.isfile('pipeline.json')
        with open('pipeline.json', 'r') as f:
            template = json.load(f)
            # The actual contents are tested in the unit
            # tests.  Just a sanity check that it looks right.
            assert "AWSTemplateFormatVersion" in template
            assert "Outputs" in template


def test_no_errors_if_override_codebuild_image(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.generate_pipeline,
            ['-i', 'python:3.6.1', 'pipeline.json'])
        assert result.exit_code == 0, result.output
        assert os.path.isfile('pipeline.json')
        with open('pipeline.json', 'r') as f:
            template = json.load(f)
            # The actual contents are tested in the unit
            # tests.  Just a sanity check that it looks right.
            image = template['Parameters']['CodeBuildImage']['Default']
            assert image == 'python:3.6.1'


def test_can_configure_github(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        # The -i option is provided so we don't have to skip this
        # test on python3.6
        result = _run_cli_command(
            runner, cli.generate_pipeline,
            ['--source', 'github', '-i' 'python:3.6.1', 'pipeline.json'])
        assert result.exit_code == 0, result.output
        assert os.path.isfile('pipeline.json')
        with open('pipeline.json', 'r') as f:
            template = json.load(f)
            # The template is already tested in the unit tests
            # for template generation.  We just want a basic
            # sanity check to make sure things are mapped
            # properly.
            assert 'GithubOwner' in template['Parameters']
            assert 'GithubRepoName' in template['Parameters']


def test_can_extract_buildspec_yaml(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.generate_pipeline,
            ['--buildspec-file', 'buildspec.yml',
             '-i', 'python:3.6.1',
             'pipeline.json'])
        assert result.exit_code == 0, result.output
        assert os.path.isfile('buildspec.yml')
        with open('buildspec.yml') as f:
            data = f.read()
            # The contents of this file are tested elsewhere,
            # we just want a basic sanity check here.
            assert 'chalice package' in data


def test_can_specify_profile_for_logs(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.logs, ['--profile', 'my-profile'],
            cli_factory=mock_cli_factory
        )
        assert result.exit_code == 0
        assert mock_cli_factory.profile == 'my-profile'


def test_can_provide_lambda_name_for_logs(runner, mock_cli_factory):
    deployed_resources = DeployedResources({
        "resources": [
            {"name": "foo",
             "lambda_arn": "arn:aws:lambda::app-dev-foo",
             "resource_type": "lambda_function"}]
    })
    mock_cli_factory.create_config_obj.return_value = FakeConfig(
        deployed_resources)
    log_retriever = mock.Mock(spec=LogRetriever)
    log_retriever.retrieve_logs.return_value = []
    mock_cli_factory.create_log_retriever.return_value = log_retriever
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.logs, ['--name', 'foo'],
            cli_factory=mock_cli_factory
        )
        assert result.exit_code == 0
    log_retriever.retrieve_logs.assert_called_with(
        LogRetrieveOptions(
            include_lambda_messages=False, max_entries=None)
    )
    mock_cli_factory.create_log_retriever.assert_called_with(
        mock.sentinel.Session, 'arn:aws:lambda::app-dev-foo', False
    )


def test_can_follow_logs_with_option(runner, mock_cli_factory):
    deployed_resources = DeployedResources({
        "resources": [
            {"name": "foo",
             "lambda_arn": "arn:aws:lambda::app-dev-foo",
             "resource_type": "lambda_function"}]
    })
    mock_cli_factory.create_config_obj.return_value = FakeConfig(
        deployed_resources)
    log_retriever = mock.Mock(spec=LogRetriever)
    log_retriever.retrieve_logs.return_value = []
    mock_cli_factory.create_log_retriever.return_value = log_retriever
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(
            runner, cli.logs, ['--name', 'foo', '--follow'],
            cli_factory=mock_cli_factory
        )
        assert result.exit_code == 0
    log_retriever.retrieve_logs.assert_called_with(
        LogRetrieveOptions(
            include_lambda_messages=False, max_entries=None)
    )
    mock_cli_factory.create_log_retriever.assert_called_with(
        mock.sentinel.Session, 'arn:aws:lambda::app-dev-foo', True
    )


def test_can_call_invoke(runner, mock_cli_factory, monkeypatch):
    invoke_handler = mock.Mock(spec=LambdaInvokeHandler)
    mock_cli_factory.create_lambda_invoke_handler.return_value = invoke_handler
    mock_reader = mock.Mock(spec=PipeReader)
    mock_reader.read.return_value = 'barbaz'
    mock_cli_factory.create_stdin_reader.return_value = mock_reader
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.invoke, ['-n', 'foo'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 0
    assert invoke_handler.invoke.call_args == mock.call('barbaz')


def test_invoke_does_raise_if_service_error(runner, mock_cli_factory):
    deployed_resources = DeployedResources({"resources": []})
    mock_cli_factory.create_config_obj.return_value = FakeConfig(
        deployed_resources)
    invoke_handler = mock.Mock(spec=LambdaInvokeHandler)
    invoke_handler.invoke.side_effect = ClientError(
        {
            'Error': {
                'Code': 'LambdaError',
                'Message': 'Error message'
            }
        },
        'Invoke'
    )
    mock_cli_factory.create_lambda_invoke_handler.return_value = invoke_handler
    mock_reader = mock.Mock(spec=PipeReader)
    mock_reader.read.return_value = 'barbaz'
    mock_cli_factory.create_stdin_reader.return_value = mock_reader
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.invoke, ['-n', 'foo'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 1
    assert invoke_handler.invoke.call_args == mock.call('barbaz')
    assert (
        "Error: got 'LambdaError' exception back from Lambda\n"
        "Error message"
    ) in result.output


def test_invoke_does_raise_if_unhandled_error(runner, mock_cli_factory):
    deployed_resources = DeployedResources({"resources": []})
    mock_cli_factory.create_config_obj.return_value = FakeConfig(
        deployed_resources)
    invoke_handler = mock.Mock(spec=LambdaInvokeHandler)
    invoke_handler.invoke.side_effect = UnhandledLambdaError('foo')
    mock_cli_factory.create_lambda_invoke_handler.return_value = invoke_handler
    mock_reader = mock.Mock(spec=PipeReader)
    mock_reader.read.return_value = 'barbaz'
    mock_cli_factory.create_stdin_reader.return_value = mock_reader
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.invoke, ['-n', 'foo'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 1
    assert invoke_handler.invoke.call_args == mock.call('barbaz')
    assert 'Unhandled exception in Lambda function, details above.' \
        in result.output


def test_invoke_does_raise_if_read_timeout(runner, mock_cli_factory):
    mock_cli_factory.create_lambda_invoke_handler.side_effect = \
        ReadTimeout('It took too long')
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.invoke, ['-n', 'foo'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 1
        assert 'It took too long' in result.output


def test_invoke_does_raise_if_no_function_found(runner, mock_cli_factory):
    mock_cli_factory.create_lambda_invoke_handler.side_effect = \
        factory.NoSuchFunctionError('foo')
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.invoke, ['-n', 'foo'],
                                  cli_factory=mock_cli_factory)
        assert result.exit_code == 2
        assert 'foo' in result.output


def test_error_message_displayed_when_missing_feature_opt_in(runner):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        with open(os.path.join('testproject', 'app.py'), 'w') as f:
            # Rather than pick an existing experimental feature, we're
            # manually injecting a feature flag into our app.  This ensures
            # we don't have to update this test if a feature graduates
            # from trial to accepted.  The '_features_used' is a "package
            # private" var for chalice code.
            f.write(
                'from chalice import Chalice\n'
                'app = Chalice("myapp")\n'
                'app._features_used.add("MYTESTFEATURE")\n'
            )
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.package, ['out'])
        assert isinstance(result.exception, ExperimentalFeatureError)
        assert 'MYTESTFEATURE' in str(result.exception)


@pytest.mark.parametrize(
    "path",
    [
        None,
        '.',
        os.getcwd,
    ],
)
def test_cli_with_absolute_path(runner, path):
    with runner.isolated_filesystem():
        if callable(path):
            path = path()
        result = runner.invoke(
            cli.cli,
            ['--project-dir', path, 'new-project', 'testproject'],
            obj={})
        assert result.exit_code == 0
        assert os.listdir(os.getcwd()) == ['testproject']
        assert_chalice_app_structure_created(dirname='testproject')


def test_can_generate_dev_plan(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.plan, [],
                                  cli_factory=mock_cli_factory)
        deployer = mock_cli_factory.create_plan_only_deployer.return_value
        call_args = deployer.deploy.call_args
        assert result.exit_code == 0
        assert isinstance(call_args[0][0], Config)
        assert call_args[1] == {'chalice_stage_name': 'dev'}


# The appgraph command actually works on py27, but due to a bug in click's
# testing (https://github.com/pallets/click/issues/848), it assumes
# stdout must be ascii.
# stdout is a cStringIO.StringIO, which doesn't accept unicode.
# See: https://docs.python.org/2/library/stringio.html#cStringIO.StringIO
@pytest.mark.skipif(sys.version_info[0] == 2,
                    reason="Click bug when writing unicode to stdout.")
def test_can_generate_appgraph(runner, mock_cli_factory):
    with runner.isolated_filesystem():
        cli.create_new_project_skeleton('testproject')
        os.chdir('testproject')
        result = _run_cli_command(runner, cli.appgraph, [])
        assert result.exit_code == 0
        # Just sanity checking some of the output
        assert 'Application' in result.output
        assert 'RestAPI(' in result.output


def test_chalice_cli_mode_env_var_always_set(runner):
    with runner.isolated_filesystem():
        result = runner.invoke(cli.new_project, ['testproject'])
        assert result.exit_code == 0
        assert os.environ['AWS_CHALICE_CLI_MODE'] == 'true'



import time
import mock
from datetime import datetime, timedelta

from chalice import logs
from chalice.awsclient import TypedAWSClient
from six import StringIO


NO_OPTIONS = logs.LogRetrieveOptions()


def message(log_message, log_stream_name='logStreamName'):
    return {
        'logStreamName': log_stream_name,
        'message': log_message,
    }


def test_can_convert_since_to_start_time():
    options = logs.LogRetrieveOptions.create(
        follow=True, since='2020-01-01T00:00:00',
        include_lambda_messages=False)
    assert options.max_entries is None
    assert options.start_time == datetime(2020, 1, 1, 0, 0, 0)
    assert not options.include_lambda_messages


def test_can_retrieve_all_logs():
    client = mock.Mock(spec=TypedAWSClient)
    log_message = message('first')
    client.iter_log_events.return_value = [log_message]
    retriever = logs.LogRetriever(client, 'loggroup')
    messages = list(retriever.retrieve_logs(NO_OPTIONS))
    expected = log_message.copy()
    # We also inject a logShortId.
    expected['logShortId'] = 'logStreamName'
    assert messages == [expected]


def test_can_support_max_entries():
    client = mock.Mock(spec=TypedAWSClient)
    client.iter_log_events.return_value = [message('first'), message('second')]
    retriever = logs.LogRetriever(client, 'loggroup')
    messages = list(
        retriever.retrieve_logs(logs.LogRetrieveOptions(max_entries=1)))
    assert len(messages) == 1
    assert messages[0]['message'] == 'first'


def test_can_exclude_lambda_messages():
    client = mock.Mock(spec=TypedAWSClient)
    client.iter_log_events.return_value = [
        message('START RequestId: id Version: $LATEST'),
        message('END RequestId: id'),
        message('REPORT RequestId: id Duration: 0.42 ms   '
                'Billed Duration: 100 ms     '
                'Memory Size: 128 MB Max Memory Used: 19 MB'),
        message('Not a lambda message'),
    ]
    retriever = logs.LogRetriever(client, 'loggroup')
    messages = list(retriever.retrieve_logs(
        logs.LogRetrieveOptions(include_lambda_messages=False)))
    assert len(messages) == 1
    assert messages[0]['message'] == 'Not a lambda message'


def test_can_parse_short_id():
    log_message = message(
        'Log Message',
        '2017/04/28/[$LATEST]fc219a0d613b40e9b5c58e6b8fd2320c'
    )
    client = mock.Mock(spec=TypedAWSClient)
    client.iter_log_events.return_value = [log_message]
    retriever = logs.LogRetriever(client, 'loggroup')
    messages = list(retriever.retrieve_logs(
        logs.LogRetrieveOptions(include_lambda_messages=False)))
    assert len(messages) == 1
    assert messages[0]['logShortId'] == 'fc219a'


def test_can_create_from_arn():
    retriever = logs.LogRetriever.create_from_lambda_arn(
        mock.sentinel.client,
        'arn:aws:lambda:us-east-1:123:function:my-function'
    )
    assert isinstance(retriever, logs.LogRetriever)


def test_can_display_logs():
    retriever = mock.Mock(spec=logs.LogRetriever)
    retriever.retrieve_logs.return_value = [
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'One'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Two'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Three'},
    ]
    stream = StringIO()
    logs.display_logs(retriever, retrieve_options=NO_OPTIONS, stream=stream)
    assert stream.getvalue().splitlines() == [
        'NOW shortId One',
        'NOW shortId Two',
        'NOW shortId Three',
    ]


def test_can_iterate_through_all_log_events():
    client = mock.Mock(spec=TypedAWSClient)
    client.iter_log_events.return_value = [
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'One'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Two'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Three'},
    ]
    event_gen = logs.LogEventGenerator(client)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=NO_OPTIONS)) == [
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'One'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Two'},
        {'timestamp': 'NOW', 'logShortId': 'shortId', 'message': 'Three'},
    ]


def test_can_follow_log_events():
    sleep = mock.Mock(spec=time.sleep)
    client = mock.Mock(spec=TypedAWSClient)
    client.filter_log_events.side_effect = [
        # First page of results has nextToken indicating there's
        # more results.
        {'events': [{'eventId': '1', 'timestamp': 1},
                    {'eventId': '2', 'timestamp': 2},
                    {'eventId': '3', 'timestamp': 3}],
         'nextToken': 'nextToken1'},
        # Second page with no more results, also note the
        # timestamps are out of order.
        {'events': [{'eventId': '4', 'timestamp': 4},
                    {'eventId': '6', 'timestamp': 6},
                    {'eventId': '5', 'timestamp': 5}]},
        # We then poll again with no new results for timestamp=6.
        {'events': [{'eventId': '6', 'timestamp': 6}]},
        # And now we get new results.
        {'events': [{'eventId': '6', 'timestamp': 6},
                    # Same timestamp we're querying (6) but a new event.
                    {'eventId': '6NEW', 'timestamp': 6},
                    {'eventId': '7', 'timestamp': 7},
                    {'eventId': '8', 'timestamp': 8}]},
        KeyboardInterrupt(),
    ]
    event_gen = logs.FollowLogEventGenerator(client, sleep)
    options = logs.LogRetrieveOptions(start_time=1)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=options)) == [
        {'eventId': '1', 'timestamp': 1},
        {'eventId': '2', 'timestamp': 2},
        {'eventId': '3', 'timestamp': 3},
        {'eventId': '4', 'timestamp': 4},
        # Note we don't try to sort these entries.
        {'eventId': '6', 'timestamp': 6},
        {'eventId': '5', 'timestamp': 5},
        {'eventId': '6NEW', 'timestamp': 6},
        {'eventId': '7', 'timestamp': 7},
        {'eventId': '8', 'timestamp': 8},
    ]
    assert client.filter_log_events.call_args_list == [
        mock.call(log_group_name='mygroup', start_time=1),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken1'),
        mock.call(log_group_name='mygroup', start_time=6),
        mock.call(log_group_name='mygroup', start_time=6),
        mock.call(log_group_name='mygroup', start_time=8),
    ]


def test_follow_logs_initially_empty():
    sleep = mock.Mock(spec=time.sleep)
    client = mock.Mock(spec=TypedAWSClient)
    client.filter_log_events.side_effect = [
        {'events': []},
        {'events': []},
        {'events': [{'eventId': '1', 'timestamp': 1},
                    {'eventId': '2', 'timestamp': 2},
                    {'eventId': '3', 'timestamp': 3}]},
        KeyboardInterrupt(),
    ]
    event_gen = logs.FollowLogEventGenerator(client, sleep)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=NO_OPTIONS)) == [
        {'eventId': '1', 'timestamp': 1},
        {'eventId': '2', 'timestamp': 2},
        {'eventId': '3', 'timestamp': 3},
    ]


def test_follow_logs_single_pages_only():
    sleep = mock.Mock(spec=time.sleep)
    client = mock.Mock(spec=TypedAWSClient)
    client.filter_log_events.side_effect = [
        {'events': [{'eventId': '1', 'timestamp': 1}]},
        {'events': [{'eventId': '2', 'timestamp': 2}]},
        {'events': [{'eventId': '3', 'timestamp': 3}]},
        KeyboardInterrupt(),
    ]
    event_gen = logs.FollowLogEventGenerator(client, sleep)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=NO_OPTIONS)) == [
        {'eventId': '1', 'timestamp': 1},
        {'eventId': '2', 'timestamp': 2},
        {'eventId': '3', 'timestamp': 3},
    ]


def test_follow_logs_last_page_empty():
    sleep = mock.Mock(spec=time.sleep)
    client = mock.Mock(spec=TypedAWSClient)
    client.filter_log_events.side_effect = [
        {'events': [{'eventId': '1', 'timestamp': 1},
                    {'eventId': '2', 'timestamp': 2},
                    {'eventId': '3', 'timestamp': 3}],
         'nextToken': 'nextToken1'},
        {'events': [{'eventId': '4', 'timestamp': 4},
                    {'eventId': '6', 'timestamp': 6},
                    {'eventId': '5', 'timestamp': 5}],
         'nextToken': 'nextToken2'},
        # You can sometimes get a next token but with no events.
        {'events': [], 'nextToken': 'nextToken3'},
        {'events': []},
        {'events': [{'eventId': '7', 'timestamp': 7}]},
        KeyboardInterrupt(),
    ]
    event_gen = logs.FollowLogEventGenerator(client, sleep)
    options = logs.LogRetrieveOptions(start_time=1)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=options)) == [
        {'eventId': '1', 'timestamp': 1},
        {'eventId': '2', 'timestamp': 2},
        {'eventId': '3', 'timestamp': 3},
        {'eventId': '4', 'timestamp': 4},
        {'eventId': '6', 'timestamp': 6},
        {'eventId': '5', 'timestamp': 5},
        {'eventId': '7', 'timestamp': 7},
    ]
    assert client.filter_log_events.call_args_list == [
        mock.call(log_group_name='mygroup', start_time=1),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken1'),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken2'),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken3'),
        mock.call(log_group_name='mygroup', start_time=6),
        mock.call(log_group_name='mygroup', start_time=7),
    ]


def test_follow_logs_all_pages_empty_with_pagination():
    sleep = mock.Mock(spec=time.sleep)
    client = mock.Mock(spec=TypedAWSClient)
    client.filter_log_events.side_effect = [
        {'events': [], 'nextToken': 'nextToken1'},
        {'events': [], 'nextToken': 'nextToken2'},
        {'events': [], 'nextToken': 'nextToken3'},
        {'events': []},
        KeyboardInterrupt(),
    ]
    event_gen = logs.FollowLogEventGenerator(client, sleep)
    options = logs.LogRetrieveOptions(start_time=1)
    assert list(event_gen.iter_log_events(
        log_group_name='mygroup', options=options)) == []
    assert client.filter_log_events.call_args_list == [
        mock.call(log_group_name='mygroup', start_time=1),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken1'),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken2'),
        mock.call(log_group_name='mygroup', start_time=1,
                  next_token='nextToken3'),
        # The last call should not use a next token.
        mock.call(log_group_name='mygroup', start_time=1)
    ]


def test_follow_logs_defaults_to_ten_minutes():
    # To avoid having to patch out/pass in utcnow(), we'll just make sure
    # that the start_time used is more recent than 10 minutes from now.
    # This is a safe assumption because we're saving the current time before
    # we invoke iter_log_events().
    ten_minutes = datetime.utcnow() - timedelta(minutes=10)
    options = logs.LogRetrieveOptions.create(follow=True)
    assert options.start_time >= ten_minutes


def test_dont_default_if_explicit_since_is_provided():
    utcnow = datetime.utcnow()
    options = logs.LogRetrieveOptions.create(follow=True, since=str(utcnow))
    assert options.start_time == utcnow



from collections import OrderedDict

import pytest

from chalice.awsclient import TypedAWSClient


@pytest.mark.parametrize('service,region,endpoint', [
    ('sns', 'us-east-1',
     OrderedDict([('partition', 'aws'),
                  ('endpointName', 'us-east-1'),
                  ('protocols', ['http', 'https']),
                  ('hostname', 'sns.us-east-1.amazonaws.com'),
                  ('signatureVersions', ['v4']),
                  ('dnsSuffix', 'amazonaws.com')])),
    ('sqs', 'cn-north-1',
     OrderedDict([('partition', 'aws-cn'),
                  ('endpointName', 'cn-north-1'),
                  ('protocols', ['http', 'https']),
                  ('sslCommonName', 'cn-north-1.queue.amazonaws.com.cn'),
                  ('hostname', 'sqs.cn-north-1.amazonaws.com.cn'),
                  ('signatureVersions', ['v4']),
                  ('dnsSuffix', 'amazonaws.com.cn')])),
    ('dynamodb', 'mars-west-1', None)
])
def test_resolve_endpoint(stubbed_session, service, region, endpoint):
    awsclient = TypedAWSClient(stubbed_session)
    assert endpoint == awsclient.resolve_endpoint(service, region)


@pytest.mark.parametrize('arn,endpoint', [
    ('arn:aws:sns:us-east-1:123456:MyTopic',
     OrderedDict([('partition', 'aws'),
                  ('endpointName', 'us-east-1'),
                  ('protocols', ['http', 'https']),
                  ('hostname', 'sns.us-east-1.amazonaws.com'),
                  ('signatureVersions', ['v4']),
                  ('dnsSuffix', 'amazonaws.com')])),
    ('arn:aws-cn:sqs:cn-north-1:444455556666:queue1',
     OrderedDict([('partition', 'aws-cn'),
                  ('endpointName', 'cn-north-1'),
                  ('protocols', ['http', 'https']),
                  ('sslCommonName', 'cn-north-1.queue.amazonaws.com.cn'),
                  ('hostname', 'sqs.cn-north-1.amazonaws.com.cn'),
                  ('signatureVersions', ['v4']),
                  ('dnsSuffix', 'amazonaws.com.cn')])),
    ('arn:aws:dynamodb:mars-west-1:123456:table/MyTable', None)
])
def test_endpoint_from_arn(stubbed_session, arn, endpoint):
    awsclient = TypedAWSClient(stubbed_session)
    assert endpoint == awsclient.endpoint_from_arn(arn)


@pytest.mark.parametrize('service,region,dns_suffix', [
    ('sns', 'us-east-1', 'amazonaws.com'),
    ('sns', 'cn-north-1', 'amazonaws.com.cn'),
    ('dynamodb', 'mars-west-1', 'amazonaws.com')
])
def test_endpoint_dns_suffix(stubbed_session, service, region, dns_suffix):
    awsclient = TypedAWSClient(stubbed_session)
    assert dns_suffix == awsclient.endpoint_dns_suffix(service, region)


@pytest.mark.parametrize('arn,dns_suffix', [
    ('arn:aws:sns:us-east-1:123456:MyTopic', 'amazonaws.com'),
    ('arn:aws-cn:sqs:cn-north-1:444455556666:queue1', 'amazonaws.com.cn'),
    ('arn:aws:dynamodb:mars-west-1:123456:table/MyTable', 'amazonaws.com')
])
def test_endpoint_dns_suffix_from_arn(stubbed_session, arn, dns_suffix):
    awsclient = TypedAWSClient(stubbed_session)
    assert dns_suffix == awsclient.endpoint_dns_suffix_from_arn(arn)


class TestServicePrincipal(object):

    @pytest.fixture
    def region(self):
        return 'bermuda-triangle-42'

    @pytest.fixture
    def url_suffix(self):
        return '.nowhere.null'

    @pytest.fixture
    def non_iso_suffixes(self):
        return ['', '.amazonaws.com', '.amazonaws.com.cn']

    @pytest.fixture
    def awsclient(self, stubbed_session):
        return TypedAWSClient(stubbed_session)

    def test_unmatched_service(self, awsclient):
        assert awsclient.service_principal('taco.magic.food.com',
                                           'us-east-1',
                                           'amazonaws.com') == \
               'taco.magic.food.com'

    def test_defaults(self, awsclient):
        assert awsclient.service_principal('lambda') == 'lambda.amazonaws.com'

    def test_states(self, awsclient, region, url_suffix, non_iso_suffixes):
        services = ['states']
        for suffix in non_iso_suffixes:
            for service in services:
                assert awsclient.service_principal('{}{}'.format(service,
                                                                 suffix),
                                                   region, url_suffix) == \
                       '{}.{}.amazonaws.com'.format(service, region)

    def test_codedeploy_and_logs(self, awsclient, region, url_suffix,
                                 non_iso_suffixes):
        services = ['codedeploy', 'logs']
        for suffix in non_iso_suffixes:
            for service in services:
                assert awsclient.service_principal('{}{}'.format(service,
                                                                 suffix),
                                                   region, url_suffix) == \
                       '{}.{}.{}'.format(service, region, url_suffix)

    def test_ec2(self, awsclient, region, url_suffix, non_iso_suffixes):
        services = ['ec2']
        for suffix in non_iso_suffixes:
            for service in services:
                assert awsclient.service_principal('{}{}'.format(service,
                                                                 suffix),
                                                   region, url_suffix) == \
                       '{}.{}'.format(service, url_suffix)

    def test_others(self, awsclient, region, url_suffix, non_iso_suffixes):
        services = ['autoscaling', 'lambda', 'events', 'sns', 'sqs',
                    'foo-service']
        for suffix in non_iso_suffixes:
            for service in services:
                assert awsclient.service_principal('{}{}'.format(service,
                                                                 suffix),
                                                   region, url_suffix) == \
                       '{}.amazonaws.com'.format(service)

    def test_local_suffix(self, awsclient, region, url_suffix):
        assert awsclient.service_principal('foo-service.local',
                                           region,
                                           url_suffix) == 'foo-service.local'

    def test_states_iso(self, awsclient):
        assert awsclient.service_principal('states.amazonaws.com',
                                           'us-iso-east-1',
                                           'c2s.ic.gov') == \
               'states.amazonaws.com'

    def test_states_isob(self, awsclient):
        assert awsclient.service_principal('states.amazonaws.com',
                                           'us-isob-east-1',
                                           'sc2s.sgov.gov') == \
               'states.amazonaws.com'

    def test_iso_exceptions(self, awsclient):
        services = ['cloudhsm', 'config', 'workspaces']
        for service in services:
            assert awsclient.service_principal(
                '{}.amazonaws.com'.format(service),
                'us-iso-east-1',
                'c2s.ic.gov') == '{}.c2s.ic.gov'.format(service)
